# -*- coding: utf-8 -*-
""" Test the main entry point """

import pytest
from py_gql.exc import SchemaError
from py_gql._graphql import graphql
from py_gql.schema import String, Schema


def test_it_correctly_identifies_r2_d2_as_the_hero_of_the_star_wars_saga(
    starwars_schema
):
    result = graphql(
        starwars_schema,
        """
        query HeroNameQuery {
            hero {
            name
            }
        }
        """,
    )
    assert result.response() == {"data": {"hero": {"name": "R2-D2"}}}


def test_correct_response_on_syntax_error_1(starwars_schema):
    assert graphql(starwars_schema, "", {}).response() == {
        "errors": [
            {
                "message": "Unexpected <EOF> (1:1):\n  1:\n    ^",
                "locations": [{"columne": 1, "line": 1}],
            }
        ]
    }


def test_correct_response_on_syntax_error_2(starwars_schema):
    query = """
    query HeroNameQuery {{
        hero {
           name
        }
    }
    """

    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [
            {
                "message": """Expected Name but found "{" (2:26):
  1:
  2:    query HeroNameQuery {{
                             ^
  3:        hero {
  4:           name""",
                "locations": [{"columne": 26, "line": 2}],
            }
        ]
    }


def test_correct_response_on_validation_errors(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery($hero: Droid) {
        hero {
            id
            foo
            friends {
                name
            }
        }
    }

    fragment hero on Character {
        id
        friends { name }
    }
    """
    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [
            {
                "locations": [{"column": 35, "line": 2}],
                "message": 'Variable "$hero" must be input type',
            },
            {
                "locations": [{"column": 13, "line": 5}],
                "message": 'Cannot query field "foo" on type "Character"',
            },
            {"message": 'Unused fragment(s) "hero"'},
            {
                "locations": [{"column": 35, "line": 2}],
                "message": 'Unused variable "$hero"',
            },
        ]
    }


def test_correct_response_on_argument_validation_error(starwars_schema):
    query = """
    query HeroNameQuery {
        luke: human {
            name
        }
    }
    """
    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [
            {
                "message": (
                    'Field "human" argument "id" of type String! '
                    "is required but not provided"
                ),
                "locations": [{"line": 3, "column": 9}],
            }
        ]
    }


def test_correct_response_on_execution_error(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery {
        hero {
            id
            friends {
                name
            }
        }
    }

    query HeroNameQuery {
        hero {
           name
        }
    }
    """
    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [
            {
                "message": "Operation name is required when document contains "
                "multiple operation definitions"
            }
        ],
        "data": None,
    }


def test_correct_response_on_execution_error_2(starwars_schema):
    query = """
    query HeroNameAndFriendsQuery {
        hero {
            id
            friends {
                name
            }
        }
    }

    query HeroNameQuery {
        hero {
           name
        }
    }
    """
    assert graphql(
        starwars_schema, query, {}, operation_name="Foo"
    ).response() == {
        "errors": [{"message": 'No operation "Foo" found in document'}],
        "data": None,
    }


def test_correct_response_on_execution_error_3(starwars_schema):
    query = """
    mutation  {
        hero {
            id
            friends {
                name
            }
        }
    }
    """
    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [{"message": "Schema doesn't support mutation operation"}],
        "data": None,
    }


def test_correct_response_on_variables_error(starwars_schema):
    query = """
    query ($episode: Episode!, $human: String!) {
        hero(episode: $episode) {
            name
        }
        human(id: $human) {
            name
        }
    }
    """
    assert graphql(
        starwars_schema, query, {"episode": 42, "id": 42}
    ).response() == {
        "errors": [
            {
                "message": (
                    'Variable "$episode" got invalid value 42 '
                    "(Expected type Episode)"
                ),
                "locations": [{"line": 2, "column": 12}],
            },
            {
                "message": (
                    'Variable "$human" of required type "String!" '
                    "was not provided."
                ),
                "locations": [{"line": 2, "column": 32}],
            },
        ],
        "data": None,
    }


def test_correct_response_on_resolver_error(starwars_schema):
    query = """
    query HeroNameQuery {
        mainHero: hero {
            name
            story: secretBackstory
        }
    }
    """
    assert graphql(starwars_schema, query, {}).response() == {
        "errors": [
            {
                "message": "secretBackstory is secret.",
                "locations": [{"line": 5, "column": 13}],
                "path": ["mainHero", "story"],
                "extensions": {"code": 42},
            }
        ],
        "data": {"mainHero": {"name": "R2-D2", "story": None}},
    }


def test_raises_if_invalid_schema_is_provided():
    with pytest.raises(SchemaError) as exc_info:
        graphql(Schema(String), "{ field }", {})
    assert str(exc_info.value) == 'Query must be ObjectType but got "String"'



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

from py_gql.validation.rules import KnownArgumentNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_single_arg_is_known(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment argOnRequiredArg on Dog {
            doesKnowCommand(dogCommand: SIT)
        }
        """,
        [],
        [],
    )


def test_multiple_args_are_known(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment multipleArgs on ComplicatedArgs {
            multipleReqs(req1: 1, req2: 2)
        }
        """,
        [],
        [],
    )


def test_ignores_args_of_unknown_fields(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment argOnUnknownField on Dog {
            unknownField(unknownArg: SIT)
        }
        """,
        [],
        [],
    )


def test_multiple_args_in_reverse_order_are_known(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment multipleArgsReverseOrder on ComplicatedArgs {
        multipleReqs(req2: 2, req1: 1)
        }
        """,
        [],
        [],
    )


def test_no_args_on_optional_arg(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment noArgOnOptionalArg on Dog {
            isHousetrained
        }
        """,
        [],
        [],
    )


def test_args_are_known_deeply(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        {
            dog {
                doesKnowCommand(dogCommand: SIT)
            }
            human {
                pet {
                    ... on Dog {
                        doesKnowCommand(dogCommand: SIT)
                    }
                }
            }
        }
        """,
        [],
        [],
    )


def test_directive_args_are_known(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        {
            dog @skip(if: true)
        }
        """,
        [],
        [],
    )


def test_unknown_directive_args_are_invalid(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
    {
        dog @skip(unless: true)
    }
    """,
        ['Unknown argument "unless" on directive "@skip"'],
        [(25, 37)],
    )


def test_misspelled_directive_args_are_reported(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
    {
        dog @skip(iff: true)
    }
    """,
        ['Unknown argument "iff" on directive "@skip", did you mean "if"?'],
        [(25, 34)],
    )


def test_invalid_arg_name(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
    fragment invalidArgName on Dog {
        doesKnowCommand(unknown: true)
    }
    """,
        ['Unknown argument "unknown" on field "doesKnowCommand" of type "Dog"'],
        [(62, 75)],
    )


def test_misspelled_arg_name_is_reported(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment invalidArgName on Dog {
            doesKnowCommand(dogcommand: true)
        }
        """,
        [
            'Unknown argument "dogcommand" on field "doesKnowCommand" of type '
            '"Dog", did you mean "dogCommand"?'
        ],
    )


def test_unknown_args_amongst_known_args(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        fragment oneGoodArgOneInvalidArg on Dog {
            doesKnowCommand(whoknows: 1, dogCommand: SIT, unknown: true)
        }
        """,
        [
            'Unknown argument "whoknows" on field "doesKnowCommand" of type "Dog"',
            'Unknown argument "unknown" on field "doesKnowCommand" of type "Dog"',
        ],
    )


def test_unknown_args_deeply(schema):
    run_test(
        KnownArgumentNamesChecker,
        schema,
        """
        {
        dog {
            doesKnowCommand(unknown: true)
        }
        human {
            pet {
                ... on Dog {
                    doesKnowCommand(unknown: true)
                }
            }
        }
        }
        """,
        [
            'Unknown argument "unknown" on field "doesKnowCommand" of type "Dog"',
            'Unknown argument "unknown" on field "doesKnowCommand" of type "Dog"',
        ],
    )



# -*- coding: utf-8 -*-
""" generic execution tests """

import json

import pytest

from py_gql.exc import ExecutionError, ResolverError, VariablesCoercionError
from py_gql.execution import execute
from py_gql.lang import parse
from py_gql.schema import (
    ID,
    Argument,
    Boolean,
    Field,
    Int,
    ListType,
    NonNullType,
    ObjectType,
    Schema,
    String,
    Type,
)

from ._test_utils import TESTED_EXECUTORS, check_execution


def test_raises_if_no_schema_is_provided():
    with pytest.raises(AssertionError) as exc_info:
        execute(None, parse("{ field }"))
    assert str(exc_info.value) == "Expected Schema object"


def test_expects_document(starwars_schema):
    with pytest.raises(AssertionError) as exc_info:
        execute(starwars_schema, None)
    assert str(exc_info.value) == "Expected document"


def test_raises_on_missing_operation(starwars_schema):
    with pytest.raises(ExecutionError) as exc_info:
        check_execution(
            starwars_schema,
            """
            fragment a on Character {
                ...b
            }
            fragment b on Character {
                ...a
            }
            """,
        )

    assert "Expected at least one operation" in str(exc_info.value)


def _test_schema(field):
    if isinstance(field, Type):
        return Schema(ObjectType("Query", [Field("test", field)]))
    else:
        return Schema(ObjectType("Query", [field]))


def test_it_uses_inline_operation_if_no_name_is_provided():
    check_execution(
        _test_schema(String),
        "{ test }",
        initial_value={"test": "foo"},
        expected_data={"test": "foo"},
    )


def test_it_uses_only_operation_if_no_name_is_provided():
    check_execution(
        _test_schema(String),
        "query Example { test }",
        initial_value={"test": "foo"},
        expected_data={"test": "foo"},
    )


def test_it_uses_named_operation_if_name_is_provided():
    check_execution(
        _test_schema(String),
        "query Example1 { test } query Example2 { test }",
        initial_value={"test": "foo"},
        operation_name="Example1",
        expected_data={"test": "foo"},
    )


def test_raises_if_no_operation_is_provided():
    with pytest.raises(ExecutionError) as exc_info:
        # This is an *invalid* query, but it should be an *executable* query.
        execute(
            _test_schema(String), parse("fragment Example on Query { test }")
        )
    assert str(exc_info.value) == "Expected at least one operation"


def test_raises_if_no_operation_name_is_provided_along_multiple_operations():
    with pytest.raises(ExecutionError) as exc_info:
        execute(
            _test_schema(String),
            parse("query Example { test } query OtherExample { test }"),
        )
    assert str(exc_info.value) == (
        "Operation name is required when document contains multiple "
        "operation definitions"
    )


def test_raises_if_unknown_operation_name_is_provided():
    with pytest.raises(ExecutionError) as exc_info:
        execute(
            _test_schema(String),
            parse("query Example { test } query OtherExample { test }"),
            operation_name="Foo",
        )
    assert str(exc_info.value) == 'No operation "Foo" found in document'


def test_it_raises_if_operation_type_is_not_supported():
    with pytest.raises(ExecutionError) as exc_info:
        assert execute(
            Schema(
                mutation_type=ObjectType("Mutation", [Field("test", String)])
            ),
            parse("{ test }"),
            initial_value={"test": "foo"},
        )
    assert str(exc_info.value) == "Schema doesn't support query operation"


def test_uses_mutation_schema_for_mutation_operation(mocker):
    query = mocker.Mock(return_value="foo")
    mutation = mocker.Mock(return_value="foo")
    subscription = mocker.Mock(return_value="foo")

    def _f(resolver):
        return Field("test", String, resolve=resolver)

    schema = Schema(
        query_type=ObjectType("Query", [_f(query)]),
        mutation_type=ObjectType("Mutation", [_f(mutation)]),
        subscription_type=ObjectType("Subscription", [_f(subscription)]),
    )

    execute(schema, parse("mutation M { test }"))
    assert not query.call_count
    assert mutation.call_count == 1


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_default_resolution_looks_up_key(exe_cls, exe_kwargs):
    schema = _test_schema(String)
    root = {"test": "testValue"}

    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            "{ test }",
            initial_value=root,
            executor=executor,
            expected_data={"test": "testValue"},
            expected_errors=[],
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_default_resolution_looks_up_attribute(exe_cls, exe_kwargs):
    class TestObject(object):
        def __init__(self, value):
            self.test = value

    schema = _test_schema(String)
    root = TestObject("testValue")
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            "{ test }",
            initial_value=root,
            executor=executor,
            expected_data={"test": "testValue"},
            expected_errors=[],
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_default_resolution_evaluates_methods(exe_cls, exe_kwargs):
    class Adder(object):
        def __init__(self, value):
            self._num = value

        def test(self, args, ctx, info):
            return self._num + args["addend1"] + ctx["addend2"]

    schema = _test_schema(Field("test", Int, [Argument("addend1", Int)]))
    root = Adder(700)

    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            "{ test(addend1: 80) }",
            initial_value=root,
            context_value={"addend2": 9},
            executor=executor,
            expected_data={"test": 789},
            expected_errors=[],
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_default_resolution_evaluates_callables(exe_cls, exe_kwargs):
    data = {
        "a": lambda *_: "Apple",
        "b": lambda *_: "Banana",
        "c": lambda *_: "Cookie",
        "d": lambda *_: "Donut",
        "e": lambda *_: "Egg",
        "deep": lambda *_: data,
    }

    Fruits = ObjectType(
        "Fruits",
        [
            Field("a", String),
            Field("b", String),
            Field("c", String),
            Field("d", String),
            Field("e", String),
            Field("deep", lambda: Fruits),
        ],
    )

    schema = Schema(Fruits)

    with exe_cls(**exe_kwargs) as executor:
        data, errors = execute(
            schema,
            parse(
                """{
                    a, b, c, d, e
                    deep {
                        a
                        deep {
                            a
                        }
                    }
                }"""
            ),
            initial_value=data,
            executor=executor,
        ).result()

    assert errors == []
    assert data == {
        "a": "Apple",
        "b": "Banana",
        "c": "Cookie",
        "d": "Donut",
        "e": "Egg",
        "deep": {"a": "Apple", "deep": {"a": "Apple"}},
    }


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_merge_of_parallel_fragments(exe_cls, exe_kwargs):
    T = ObjectType(
        "Type",
        [
            Field("a", String, resolve=lambda *_: "Apple"),
            Field("b", String, resolve=lambda *_: "Banana"),
            Field("c", String, resolve=lambda *_: "Cherry"),
            Field("deep", lambda: T, resolve=lambda *_: dict()),
        ],
    )
    schema = Schema(T)

    with exe_cls(**exe_kwargs) as executor:
        data, errors = execute(
            schema,
            parse(
                """
                { a, ...FragOne, ...FragTwo }

                fragment FragOne on Type {
                    b
                    deep { b, deeper: deep { b } }
                }

                fragment FragTwo on Type {
                    c
                    deep { c, deeper: deep { c } }
                }
                """
            ),
            executor=executor,
        ).result()

    assert errors == []
    assert data == {
        "a": "Apple",
        "b": "Banana",
        "c": "Cherry",
        "deep": {
            "b": "Banana",
            "c": "Cherry",
            "deeper": {"b": "Banana", "c": "Cherry"},
        },
    }


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_forwarded_resolver_arguments(mocker, exe_cls, exe_kwargs):

    resolver, context, root = mocker.Mock(return_value="foo"), {}, {}

    field = Field("test", String, [Argument("arg", String)], resolve=resolver)
    query_type = ObjectType("Test", [field])
    doc = parse("query ($var: String) { result: test(arg: $var) }")
    schema = Schema(query_type)

    with exe_cls(**exe_kwargs) as executor:
        _, errors = execute(
            schema,
            doc,
            context_value=context,
            initial_value=root,
            variables={"var": 123},
            executor=executor,
        ).result()

        assert errors == []

        parent_value, args, ctx, info = resolver.call_args[0]

        assert info.field_def is field
        assert info.parent_type is query_type
        assert info.path == ["result"]
        assert info.variables == {"var": "123"}
        assert info.schema is schema
        assert info.operation is doc.definitions[0]
        assert info.executor is executor

        assert ctx is context
        assert parent_value is root

        assert args == {"arg": "123"}


NullNonNullDataType = ObjectType(
    "DataType",
    [
        Field("scalar", String),
        Field("scalarNonNull", NonNullType(String)),
        Field("nested", lambda: NullNonNullDataType),
        Field("nestedNonNull", lambda: NonNullType(NullNonNullDataType)),
    ],
)

NullAndNonNullSchema = Schema(NullNonNullDataType)


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_nullable_field(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            NullAndNonNullSchema,
            "query Q { scalar }",
            initial_value=dict(scalar=None),
            expected_data={"scalar": None},
            expected_errors=[],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_lazy_nullable_field(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            NullAndNonNullSchema,
            "query Q { scalar }",
            initial_value=dict(scalar=lambda *_: None),
            expected_data={"scalar": None},
            expected_errors=[],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_and_report_error_on_non_nullable_field(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            NullAndNonNullSchema,
            "query Q { scalarNonNull }",
            initial_value=dict(scalarNonNull=None),
            expected_data={"scalarNonNull": None},
            expected_errors=[
                (
                    'Field "scalarNonNull" is not nullable',
                    (10, 23),
                    "scalarNonNull",
                )
            ],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_and_report_error_on_lazy_non_nullable_field(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            NullAndNonNullSchema,
            "query Q { scalarNonNull }",
            initial_value=dict(scalarNonNull=lambda *_: None),
            expected_data={"scalarNonNull": None},
            expected_errors=[
                (
                    'Field "scalarNonNull" is not nullable',
                    (10, 23),
                    "scalarNonNull",
                )
            ],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_tree_of_nullable_fields(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            NullAndNonNullSchema,
            """
            query Q {
                nested {
                    scalar
                    nested {
                        scalar
                        nested {
                            scalar
                        }
                    }
                }
            }
            """,
            initial_value={
                "nested": {
                    "scalar": None,
                    "nested": {"scalar": None, "nested": None},
                }
            },
            expected_data={
                "nested": {
                    "nested": {"nested": None, "scalar": None},
                    "scalar": None,
                }
            },
            expected_errors=[],
            executor=executor,
        )


# Depending on scheduling the errors can be in a different order
def test_nulls_and_report_errors_on_tree_of_non_nullable_fields():

    check_execution(
        NullAndNonNullSchema,
        """
        query Q {
            nested {
                scalarNonNull
                nestedNonNull {
                    scalar
                    nestedNonNull {
                        scalarNonNull
                    }
                }
            }
            nestedNonNull {
                scalarNonNull
            }
        }
        """,
        initial_value={
            "nestedNonNull": None,
            "nested": {"scalarNonNull": None, "nestedNonNull": None},
        },
        expected_data={
            "nested": {"nestedNonNull": None, "scalarNonNull": None},
            "nestedNonNull": None,
        },
        expected_errors=[
            (
                'Field "nested.scalarNonNull" is not nullable',
                (56, 69),
                "nested.scalarNonNull",
            ),
            (
                'Field "nested.nestedNonNull" is not nullable',
                (86, 242),
                "nested.nestedNonNull",
            ),
            (
                'Field "nestedNonNull" is not nullable',
                (269, 328),
                "nestedNonNull",
            ),
        ],
    )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_nulls_out_errored_subtrees(raiser, exe_cls, exe_kwargs):
    doc = parse(
        """{
        sync,
        callable_error,
        callable,
        resolver_error,
        resolver,
    }"""
    )

    root = dict(
        sync="sync",
        callable_error=raiser(ResolverError, "callable_error"),
        callable=lambda *_: "callable",
    )

    schema = Schema(
        ObjectType(
            "Query",
            [
                Field("sync", String),
                Field("callable_error", String),
                Field("callable", String),
                Field(
                    "resolver_error",
                    String,
                    resolve=raiser(ResolverError, "resolver_error"),
                ),
                Field("resolver", String, resolve=lambda *_: "resolver"),
            ],
        )
    )

    with exe_cls(**exe_kwargs) as executor:
        data, errors = execute(
            schema, doc, initial_value=root, executor=executor
        ).result()

    assert data == {
        "sync": "sync",
        "callable_error": None,
        "callable": "callable",
        "resolver_error": None,
        "resolver": "resolver",
    }

    assert [(str(err), err.nodes[0].loc, err.path) for err in errors] == [
        ("callable_error", (24, 38), ["callable_error"]),
        ("resolver_error", (66, 80), ["resolver_error"]),
    ]


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_full_response_path_is_included_on_error(raiser, exe_cls, exe_kwargs):
    A = ObjectType(
        "A",
        [
            Field("nullableA", lambda: A, resolve=lambda *_: {}),
            Field("nonNullA", lambda: NonNullType(A), resolve=lambda *_: {}),
            Field(
                "raises",
                lambda: NonNullType(String),
                resolve=raiser(ResolverError, "Catch me if you can"),
            ),
        ],
    )
    schema = Schema(
        ObjectType(
            "query", [Field("nullableA", lambda: A, resolve=lambda *_: {})]
        )
    )

    doc = parse(
        """
    query {
        nullableA {
            aliasedA: nullableA {
                nonNullA {
                    anotherA: nonNullA {
                        raises
                    }
                }
            }
        }
    }
    """
    )

    with exe_cls(**exe_kwargs) as executor:
        data, errors = execute(schema, doc, executor=executor).result()

    assert data == {
        "nullableA": {"aliasedA": {"nonNullA": {"anotherA": {"raises": None}}}}
    }

    assert [(str(err), err.nodes[0].loc, err.path) for err in errors] == [
        (
            "Catch me if you can",
            (159, 165),
            ["nullableA", "aliasedA", "nonNullA", "anotherA", "raises"],
        )
    ]


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_it_does_not_include_illegal_fields(mocker, exe_cls, exe_kwargs):
    """ ...even if you skip validation """
    # This is an *invalid* query, but it should be an *executable* query.
    doc = parse(
        """
    mutation M {
        thisIsIllegalDontIncludeMe
    }
    """
    )

    schema = Schema(
        mutation_type=ObjectType("mutation", [Field("test", String)])
    )

    root = {
        "test": mocker.Mock(return_value="foo"),
        "thisIsIllegalDontIncludeMe": mocker.Mock(return_value="foo"),
    }

    with exe_cls(**exe_kwargs) as executor:
        result, _ = execute(
            schema, doc, initial_value=root, executor=executor
        ).result()
    assert result == {}

    assert not root["thisIsIllegalDontIncludeMe"].call_count


@pytest.fixture
def _complex_schema():
    BlogImage = ObjectType(
        "Image",
        [Field("url", String), Field("width", Int), Field("height", Int)],
    )

    BlogAuthor = ObjectType(
        "Author",
        [
            Field("id", String),
            Field("name", String),
            Field(
                "pic",
                BlogImage,
                [Argument("width", Int), Argument("height", Int)],
            ),
            Field("recentArticle", lambda: BlogArticle),
        ],
    )

    BlogArticle = ObjectType(
        "Article",
        [
            Field("id", String),
            Field("isPublished", Boolean),
            Field("author", BlogAuthor),
            Field("title", String),
            Field("body", String),
            Field("keywords", ListType(String)),
        ],
    )

    BlogQuery = ObjectType(
        "Query",
        [
            Field(
                "article",
                BlogArticle,
                [Argument("id", ID)],
                resolve=lambda _, args, *r: article(args["id"]),
            ),
            Field(
                "feed",
                ListType(BlogArticle),
                resolve=lambda *_: [article(i) for i in range(1, 11)],
            ),
        ],
    )

    def article(id_):
        return {
            "id": id_,
            "isPublished": True,
            "author": lambda *r: john_smith,
            "title": "My Article " + str(id_),
            "body": "This is a post",
            "hidden": "This data is not exposed in the schema",
            "keywords": ["foo", "bar", 1, True, None],
        }

    john_smith = {
        "id": 123,
        "name": "John Smith",
        "recentArticle": article(1),
        "pic": lambda _, args, *r: {
            "url": "cdn://123",
            "width": args["width"],
            "height": args["height"],
        },
    }

    schema = Schema(BlogQuery)

    query = """
    {
        feed {
            id,
            title
        },
        article(id: "1") {
            ...articleFields,
            author {
                id,
                name,
                pic(width: 640, height: 480) {
                    url,
                    width,
                    height
                },
                recentArticle {
                    ...articleFields,
                    keywords
                }
            }
        }
    }

    fragment articleFields on Article {
        id,
        isPublished,
        title,
        body,
        hidden,
        notdefined
    }
    """

    return schema, parse(query)


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_executes_correctly_without_validation(
    _complex_schema, exe_cls, exe_kwargs
):
    # This is an *invalid* query, but it should be an *executable* query.
    schema, doc = _complex_schema

    with exe_cls(**exe_kwargs) as executor:
        data, errors = execute(schema, doc, executor=executor).result()

    assert data == {
        "article": {
            "author": {
                "id": "123",
                "name": "John Smith",
                "pic": {"url": "cdn://123", "width": 640, "height": 480},
                "recentArticle": {
                    "id": "1",
                    "isPublished": True,
                    "title": "My Article 1",
                    "body": "This is a post",
                    "keywords": ["foo", "bar", "1", "true", None],
                },
            },
            "body": "This is a post",
            "id": "1",
            "isPublished": True,
            "title": "My Article 1",
        },
        "feed": [
            {"id": "1", "title": "My Article 1"},
            {"id": "2", "title": "My Article 2"},
            {"id": "3", "title": "My Article 3"},
            {"id": "4", "title": "My Article 4"},
            {"id": "5", "title": "My Article 5"},
            {"id": "6", "title": "My Article 6"},
            {"id": "7", "title": "My Article 7"},
            {"id": "8", "title": "My Article 8"},
            {"id": "9", "title": "My Article 9"},
            {"id": "10", "title": "My Article 10"},
        ],
    }

    assert errors == []


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_result_is_ordered_according_to_query(
    _complex_schema, exe_cls, exe_kwargs
):
    """ check that deep iteration order of keys in result corresponds to order
    of appearance in query accounting for fragment use """
    # This is an *invalid* query, but it should be an *executable* query.
    schema, doc = _complex_schema
    with exe_cls(**exe_kwargs) as executor:
        data, _ = execute(schema, doc, executor=executor).result()

    def _extract_keys_in_order(d):
        if not isinstance(d, dict):
            return None
        keys = []
        for key, value in d.items():
            if isinstance(value, dict):
                keys.append((key, _extract_keys_in_order(value)))
            elif isinstance(value, list):
                keys.append((key, [_extract_keys_in_order(i) for i in value]))
            else:
                keys.append((key, None))
        return keys

    assert _extract_keys_in_order(data) == [
        (
            "feed",
            [
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
                [("id", None), ("title", None)],
            ],
        ),
        (
            "article",
            [
                ("id", None),
                ("isPublished", None),
                ("title", None),
                ("body", None),
                (
                    "author",
                    [
                        ("id", None),
                        ("name", None),
                        (
                            "pic",
                            [("url", None), ("width", None), ("height", None)],
                        ),
                        (
                            "recentArticle",
                            [
                                ("id", None),
                                ("isPublished", None),
                                ("title", None),
                                ("body", None),
                                ("keywords", [None, None, None, None, None]),
                            ],
                        ),
                    ],
                ),
            ],
        ),
    ]


class TestNonNullArguments(object):
    schema_with_null_args = Schema(
        ObjectType(
            "Query",
            [
                Field(
                    "withNonNullArg",
                    String,
                    args=[Argument("cannotBeNull", NonNullType(String))],
                    resolve=lambda _, a, *__: json.dumps(
                        a.get("cannotBeNull", "NOT PROVIDED")
                    ),
                )
            ],
        )
    )

    def test_non_null_literal(self):
        check_execution(
            self.schema_with_null_args,
            """
            query {
                withNonNullArg (cannotBeNull: "literal value")
            }
            """,
            expected_data={"withNonNullArg": '"literal value"'},
            expected_errors=[],
        )

    def test_non_null_variable(self):
        check_execution(
            self.schema_with_null_args,
            """
            query ($testVar: String!) {
                withNonNullArg (cannotBeNull: $testVar)
            }
            """,
            variables={"testVar": "variable value"},
            expected_data={"withNonNullArg": '"variable value"'},
            expected_errors=[],
        )

    def test_missing_variable_with_default(self):
        check_execution(
            self.schema_with_null_args,
            """
            query ($testVar: String = "default value") {
                withNonNullArg (cannotBeNull: $testVar)
            }
            """,
            expected_data={"withNonNullArg": '"default value"'},
            expected_errors=[],
        )

    def test_missing(self):
        check_execution(
            self.schema_with_null_args,
            """
            query {
                withNonNullArg
            }
            """,
            expected_data={"withNonNullArg": None},
            expected_errors=[
                (
                    'Argument "cannotBeNull" of required type "String!" was '
                    "not provided",
                    (37, 51),
                    "withNonNullArg",
                )
            ],
        )

    def test_null_literal(self):
        check_execution(
            self.schema_with_null_args,
            """
            query {
                withNonNullArg (cannotBeNull: null)
            }
            """,
            expected_data={"withNonNullArg": None},
            expected_errors=[
                (
                    'Argument "cannotBeNull" of type "String!" was provided '
                    "invalid value null (Expected non null value.)",
                    (37, 72),
                    "withNonNullArg",
                )
            ],
        )

    def test_missing_variable(self):
        # Differs from reference implementation as a missing variable will
        # abort the full execution. This is consistent as all variables defined
        # must be used in an operation and so a missing variables for a non null
        # type should break.
        check_execution(
            self.schema_with_null_args,
            """
            query ($testVar: String!) {
                withNonNullArg (cannotBeNull: $testVar)
            }
            """,
            expected_exc=VariablesCoercionError,
            expected_msg=(
                'Variable "$testVar" of required type "String!" was not '
                "provided."
            ),
        )

    def test_null_variable(self):
        # Differs from reference implementation as a null variable provided for
        # a non null type will abort the full execution.
        check_execution(
            self.schema_with_null_args,
            """
            query ($testVar: String!) {
                withNonNullArg (cannotBeNull: $testVar)
            }
            """,
            variables={"testVar": None},
            expected_exc=VariablesCoercionError,
            expected_msg=(
                'Variable "$testVar" of required type "String!" '
                "must not be null."
            ),
        )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueFragmentNamesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_no_fragments(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    {
        field
    }
    """,
    )


def test_one_fragment(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    {
        ...fragA
    }

    fragment fragA on Type {
        field
    }
    """,
    )


def test_many_fragments(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    {
        ...fragA
        ...fragB
        ...fragC
    }
    fragment fragA on Type {
        fieldA
    }
    fragment fragB on Type {
        fieldB
    }
    fragment fragC on Type {
        fieldC
    }
    """,
    )


def test_inline_fragments_are_always_unique(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    {
        ...on Type {
            fieldA
        }
        ...on Type {
            fieldB
        }
    }
    """,
    )


def test_fragment_and_operation_named_the_same(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    query Foo {
        ...Foo
    }
    fragment Foo on Type {
        field
    }
    """,
    )


def test_fragments_named_the_same(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    {
        ...fragA
    }
    fragment fragA on Type {
        fieldA
    }
    fragment fragA on Type {
        fieldB
    }
    """,
        ['There can only be one fragment named "fragA"'],
    )


def test_fragments_named_the_same_without_being_referenced(schema):
    run_test(
        UniqueFragmentNamesChecker,
        schema,
        """
    fragment fragA on Type {
        fieldA
    }
    fragment fragA on Type {
        fieldB
    }
    """,
        ['There can only be one fragment named "fragA"'],
    )



# -*- coding: utf-8 -*-
""" generic mutations handling tests """

import pytest

from py_gql.exc import ResolverError
from py_gql.schema import Argument, Field, Int, ObjectType, Schema

from ._test_utils import TESTED_EXECUTORS, check_execution


class NumberHolder(object):
    __slots__ = "theNumber"

    def __init__(self, original):
        self.theNumber = original


class Root(object):
    def __init__(self, original):
        self.numberHolder = NumberHolder(original)

    def incrementTheNumber(self, steps):
        self.numberHolder.theNumber = self.numberHolder.theNumber + steps
        return self.numberHolder

    def immediatelyChangeTheNumber(self, newNumber):
        self.numberHolder.theNumber = newNumber
        return self.numberHolder

    def lazilyChangeTheNumber(self, newNumber):
        def _mutate():
            self.numberHolder.theNumber = newNumber
            return self.numberHolder

        return _mutate

    def failToChangeTheNumber(self, newNumber):
        raise ResolverError("Cannot change the number")

    def lazilyFailToChangeTheNumber(self, newNumber):
        def _mutate():
            raise ResolverError("Cannot change the number")

        return _mutate


number_holder = ObjectType("NumberHolder", [Field("theNumber", Int)])

schema = Schema(
    ObjectType("Query", [Field("numberHolder", number_holder)]),
    ObjectType(
        "Mutation",
        [
            Field(
                "incrementTheNumber",
                number_holder,
                args=[Argument("steps", Int)],
                resolve=lambda obj, args, *_: obj.incrementTheNumber(
                    args["steps"]
                ),
            ),
            Field(
                "immediatelyChangeTheNumber",
                number_holder,
                args=[Argument("newNumber", Int)],
                resolve=lambda obj, args, *_: (
                    obj.immediatelyChangeTheNumber(args["newNumber"])
                ),
            ),
            Field(
                "lazilyChangeTheNumber",
                number_holder,
                args=[Argument("newNumber", Int)],
                resolve=lambda obj, args, *_: (
                    obj.lazilyChangeTheNumber(args["newNumber"])
                ),
            ),
            Field(
                "failToChangeTheNumber",
                number_holder,
                args=[Argument("newNumber", Int)],
                resolve=lambda obj, args, *_: (
                    obj.failToChangeTheNumber(args["newNumber"])
                ),
            ),
            Field(
                "lazilyFailToChangeTheNumber",
                number_holder,
                args=[Argument("newNumber", Int)],
                resolve=lambda obj, args, *_: (
                    obj.lazilyFailToChangeTheNumber(args["newNumber"])
                ),
            ),
        ],
    ),
)


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_it_evaluates_mutations_serially(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            """
            mutation M {
                first: immediatelyChangeTheNumber(newNumber: 1) {
                    theNumber
                },
                second: lazilyChangeTheNumber(newNumber: 2) {
                    theNumber
                },
                third: immediatelyChangeTheNumber(newNumber: 3) {
                    theNumber
                }
                fourth: lazilyChangeTheNumber(newNumber: 4) {
                    theNumber
                },
                fifth: immediatelyChangeTheNumber(newNumber: 5) {
                    theNumber
                }
            }
            """,
            initial_value=Root(6),
            expected_data={
                "first": {"theNumber": 1},
                "second": {"theNumber": 2},
                "third": {"theNumber": 3},
                "fourth": {"theNumber": 4},
                "fifth": {"theNumber": 5},
            },
            expected_errors=[],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_it_evaluates_mutations_serially_2(exe_cls, exe_kwargs):
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            """
            mutation M {
                first: incrementTheNumber(steps: 1) {
                    theNumber
                },
                second: incrementTheNumber(steps: -3) {
                    theNumber
                }
            }
            """,
            initial_value=Root(6),
            expected_data={
                "first": {"theNumber": 7},
                "second": {"theNumber": 4},
            },
            expected_errors=[],
            executor=executor,
        )


@pytest.mark.parametrize("exe_cls, exe_kwargs", TESTED_EXECUTORS)
def test_it_evaluates_mutations_correctly_even_when_some_mutation_fails(
    exe_cls, exe_kwargs
):

    doc = """
        mutation M {
            first: immediatelyChangeTheNumber(newNumber: 1) {
                theNumber
            },
            second: lazilyChangeTheNumber(newNumber: 2) {
                theNumber
            },
            third: failToChangeTheNumber(newNumber: 3) {
                theNumber
            }
            fourth: lazilyChangeTheNumber(newNumber: 4) {
                theNumber
            },
            fifth: immediatelyChangeTheNumber(newNumber: 5) {
                theNumber
            }
            sixth: lazilyFailToChangeTheNumber(newNumber: 6) {
                theNumber
            }
        }
    """
    with exe_cls(**exe_kwargs) as executor:
        check_execution(
            schema,
            doc,
            initial_value=Root(6),
            executor=executor,
            expected_data={
                "first": {"theNumber": 1},
                "second": {"theNumber": 2},
                "third": None,
                "fourth": {"theNumber": 4},
                "fifth": {"theNumber": 5},
                "sixth": None,
            },
            expected_errors=[
                ("Cannot change the number", (236, 320), "third"),
                ("Cannot change the number", (534, 624), "sixth"),
            ],
        )



# -*- coding: utf-8 -*-
""" Useful tracer implementations. """

import datetime as dt
import json
import logging

from .._utils import OrderedDict
from ..execution.tracing import GraphQLTracer
from ..execution.wrappers import GraphQLExtension


def _nanoseconds(delta):
    return int(delta.total_seconds() * 1e9)


def _rfc3339(ts):
    return ts.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


class ApolloTracer(GraphQLTracer, GraphQLExtension):
    """ `Apollo Tracing <https://github.com/apollographql/apollo-tracing>`_
    implementation. This tracers also implements :class:`GraphQLExtension`
    so you can add the result to the response. """

    def __init__(self):
        self.start = None
        self.validation_start = None
        self.validation_end = None
        self.parse_start = None
        self.parse_end = None
        self.fields = OrderedDict()
        self.end = None

    def on_query_start(self, **_):
        self.start = dt.datetime.utcnow()

    def on_query_end(self, **_):
        self.end = dt.datetime.utcnow()

    def on_parse_start(self, **_):
        self.parse_start = dt.datetime.utcnow()

    def on_parse_end(self, **_):
        self.parse_end = dt.datetime.utcnow()

    def on_validate_start(self, **_):
        self.validation_start = dt.datetime.utcnow()

    def on_validate_end(self, **_):
        self.validation_end = dt.datetime.utcnow()

    def on_field_start(self, info, **_):
        start = dt.datetime.utcnow()
        self.fields[tuple(info.path)] = {
            "path": list(info.path),
            "parentType": info.parent_type.name,
            "fieldName": info.field_def.name,
            "returnType": str(info.field_def.type),
            "startOffset": _nanoseconds(start - self.start),
            "start": start,
        }

    def on_field_end(self, info, **_):
        end = dt.datetime.utcnow()
        start = self.fields[tuple(info.path)].pop("start")
        self.fields[tuple(info.path)]["duration"] = _nanoseconds(end - start)

    def name(self):
        return "tracing"

    def payload(self):
        return {
            "version": 1,
            "startTime": _rfc3339(self.start),
            "endTime": _rfc3339(self.end),
            "duration": _nanoseconds(self.end - self.start),
            "execution": (
                {"resolvers": list(self.fields.values())}
                if self.fields
                else None
            ),
            "validation": (
                {
                    "duration": _nanoseconds(
                        self.validation_end - self.validation_start
                    ),
                    "startOffset": _nanoseconds(
                        self.validation_start - self.start
                    ),
                }
                if self.validation_start is not None
                else None
            ),
            "parsing": (
                {
                    "duration": _nanoseconds(self.parse_end - self.parse_start),
                    "startOffset": _nanoseconds(self.parse_start - self.start),
                }
                if self.parse_start is not None
                else None
            ),
        }


_SLOW_LOG_FORMAT_STR = """GraphQL query took too long (duration = %fms, \
threshold = %fms, operation = %s, document = '''
%s
''', variables = %s)"""

_DEFAULT_LOGGER = logging.getLogger("py_gql.utilities.tracers.SlowQueryLog")


class SlowQueryLog(GraphQLTracer):
    """ Log slow queries through Python's logging utilities.

    Note:
        By default this logs the entire query and variables, if this is not
        suitable (e.g. you need to redact the query and or variables) you can
        subclass  and override :meth:`format_document` and
        :meth:`format_variables`.

    Args:
        threshold (int): Slow query threshold in ms

        logger (Optional[logging.Logger]): Custome logger instance to use.
            Defaults to ``py_gql.utilities.tracers.SlowQueryLog``.

        level (Optional[int]): Log level. Defaults to ``WARNING``.

        format_str (Optional[str]): Log format string.
            The log call will pass the following variables: (duration of the
            query in ms, threshold in ms, operation name if any, formatted
            document, formatted variables)
    """

    def __init__(
        self, threshold, logger=None, level=logging.WARNING, format_str=None
    ):
        self._threshold = threshold
        self._logger = logger if logger is not None else _DEFAULT_LOGGER
        self._level = level
        self._format_str = format_str or _SLOW_LOG_FORMAT_STR

    def format_document(self, document):
        return document

    def format_variables(self, variables):
        return json.dumps(variables, indent=4, sort_keys=True)

    def on_query_start(self, **_):
        self._start = dt.datetime.utcnow()

    def on_query_end(self, document, variables, operation_name):
        end = dt.datetime.utcnow()
        duration = (end - self._start).total_seconds() * 1000
        if duration > self._threshold:
            self._logger.log(
                self._level,
                self._format_str,
                duration,
                self._threshold,
                operation_name,
                self.format_document(document),
                self.format_variables(variables),
            )



# -*- coding: utf-8 -*-
""" Utilities to validate Python values against a schema / types """

import json

from .._utils import find_one
from ..exc import (
    CoercionError,
    InvalidValue,
    MultiCoercionError,
    ScalarParsingError,
    UnknownEnumValue,
    UnknownType,
    VariableCoercionError,
    VariablesCoercionError,
)
from ..lang import ast as _ast, print_ast
from ..schema import (
    EnumType,
    InputObjectType,
    ListType,
    NonNullType,
    ScalarType,
    is_input_type,
)
from .value_from_ast import value_from_ast


def _path(path):
    if not path:
        return []
    return ["value"] + path


def coerce_value(value, type_, node=None, path=None):
    """ Coerce a Python value given a GraphQL Type.

    Args:
        value (any): Value to coerce
        type_ (py_gql.schema.Type): Expected type
        node (Optional[py_gql.lang.ast.Node]): Relevant node
        path: (Optional[list]):
            Path into the value for nested values (lists, objects).
            Should only be set on recursive calls.

    Returns:
        any: The coerced value

    Raises:
        :class:`~py_gql.exc.CoercionError`: if coercion fails
    """
    if path is None:
        path = []

    if isinstance(type_, NonNullType):
        if value is None:
            raise CoercionError(
                "Expected non-nullable type %s not to be null" % type_,
                node,
                value_path=_path(path),
            )
        type_ = type_.type

    if value is None:
        return None

    if isinstance(type_, ScalarType):
        try:
            return type_.parse(value)
        except ScalarParsingError as err:
            raise CoercionError(str(err), node, value_path=_path(path))

    if isinstance(type_, EnumType):
        if isinstance(value, str):
            try:
                return type_.get_value(value)
            except UnknownEnumValue as err:
                raise CoercionError(str(err), node, value_path=_path(path))
        else:
            raise CoercionError(
                "Expected type %s" % type_, node, value_path=_path(path)
            )

    if isinstance(type_, ListType):
        return _coerce_list_value(value, type_, node, path)

    if isinstance(type_, InputObjectType):
        return _coerce_input_object(value, type_, node, path)


def _coerce_list_value(value, type_, node, path):
    if isinstance(value, (list, tuple)):
        coerced = []
        errors = []

        for index, entry in enumerate(value):
            try:
                coerced.append(
                    coerce_value(
                        entry, type_.type, node=node, path=path + [index]
                    )
                )
            except MultiCoercionError as err:
                for child_err in err.errors:
                    errors.append(child_err)
            except CoercionError as err:
                errors.append(err)

        if len(errors) > 1:
            raise MultiCoercionError(errors)
        elif len(errors) == 1:
            raise errors[0]

        return coerced
    else:
        return [coerce_value(value, type_.type, node=node, path=path + [0])]


def _coerce_input_object(value, type_, node, path):
    if not isinstance(value, dict):
        raise CoercionError(
            "Expected type %s to be an object" % type_,
            node,
            value_path=_path(path),
        )

    coerced = {}
    errors = []
    for field in type_.fields:
        if field.name not in value:
            if isinstance(field.type, NonNullType):
                errors.append(
                    CoercionError(
                        "Field %s of required type %s was not provided"
                        % (field.name, field.type),
                        node,
                        value_path=_path(path + [field.name]),
                    )
                )
        else:
            try:
                coerced[field.name] = coerce_value(
                    value[field.name], field.type, node, path + [field.name]
                )
            except MultiCoercionError as err:
                for child_err in err.errors:
                    errors.append(child_err)
            except CoercionError as err:
                errors.append(err)

    if len(errors) > 1:
        raise MultiCoercionError(errors)
    elif len(errors) == 1:
        raise errors[0]

    for fieldname in value.keys():
        if fieldname not in type_.field_map:
            raise CoercionError(
                "Field %s is not defined by type %s" % (fieldname, type_),
                node,
                value_path=_path(path),
            )

    return coerced


def coerce_argument_values(definition, node, variables=None):
    """ Prepare a dict of argument values given a field or directive
    definition and a field or directive node.

    Args:
        definition (Union[py_gql.schema.Directive, py_gql.schema.Field]):
            Field or Directive definition from which to extract argument
            definitions.
        node (Union[py_gql.lang.ast.Field, py_gql.lang.ast.Directive]):
            AST node
        variables (Optional[dict]):  Coerced variable values


    Returns:
        dict: Coerced arguments

    Raises:
        :class:`~py_gql.exc.CoercionError`:
            if any argument value fails to coerce, required argument is
            missing, etc.
    """
    variables = dict() if variables is None else variables
    coerced_values = {}

    values = {a.name.value: a for a in node.arguments}
    defs = definition.arguments
    for arg_def in defs:
        argname = arg_def.name
        argtype = arg_def.type
        if argname not in values:
            if arg_def.has_default_value:
                coerced_values[argname] = arg_def.default_value
            elif isinstance(argtype, NonNullType):
                raise CoercionError(
                    'Argument "%s" of required type "%s" was not provided'
                    % (argname, argtype),
                    [node],
                )
        else:
            arg = values[argname]
            if isinstance(arg.value, _ast.Variable):
                varname = arg.value.name.value
                if varname in variables:
                    coerced_values[argname] = variables[varname]
                elif arg_def.has_default_value:
                    coerced_values[argname] = arg_def.default_value
                elif isinstance(argtype, NonNullType):
                    raise CoercionError(
                        'Argument "%s" of required type "%s" was provided the '
                        'missing variable "$%s"' % (argname, argtype, varname),
                        [node],
                    )
            else:
                try:
                    coerced_values[argname] = value_from_ast(
                        arg.value, argtype, variables=variables
                    )
                except InvalidValue as err:
                    raise CoercionError(
                        'Argument "%s" of type "%s" was provided invalid value '
                        "%s (%s)"
                        % (argname, argtype, print_ast(arg.value), str(err)),
                        [node],
                    )

    return coerced_values


def directive_arguments(definition, node, variables=None):
    """ Extract directive argument given a field node and a directive
    definition.

    Args:
        definition (py_gql.schema.Directive):
            Directive definition from which to extract arguments
        node (py_gql.lang.ast.Field): Ast node
        variables (Optional[dict]): Coerced variable values

    Returns:
        Optional[dict]: Coerced directive arguments, ``None`` if the directive
        is not present on the node.

    Raises:
        :class:`~py_gql.exc.CoercionError`:
            if any directive argument value fails to coerce, required argument
            is missing, etc.
    """
    directive = find_one(
        node.directives, lambda d: d.name.value == definition.name
    )

    return (
        coerce_argument_values(definition, directive, variables)
        if directive is not None
        else None
    )


def coerce_variable_values(schema, operation, variables):  # noqa: C901
    """ Prepares an object map of variable values of the correct type based on
    the provided operation definition and arbitrary JSON input. If the input
    cannot be parsed to match the variable definitions, an ExecutionError will
    be thrown. The returned value is a plain dict since it is exposed to user
    code.

    Extra variables are ignored and filtered out.

    Args:
        schema (py_gql.schema.Schema): GraphQL Schema to consider
        operation (py_gql.lang.ast.OperationDefinition):
            Operation definition containing the variable definitions
        variables (dict): Provided raw variables

    Returns:
        dict: Coerced variables

    Raises:
        :class:`~py_gql.exc.VariablesCoercionError`:
            if any variable cannot be coerced.
    """
    coerced, errors = {}, []

    for var_def in operation.variable_definitions:
        name = var_def.variable.name.value

        try:
            var_type = schema.get_type_from_literal(var_def.type)
        except UnknownType as err:
            errors.append(
                VariableCoercionError(
                    'Unknown type "%s" for variable "$%s"'
                    % (print_ast(var_def.type), name),
                    [var_def],
                )
            )
            continue

        if not is_input_type(var_type):
            errors.append(
                VariableCoercionError(
                    'Variable "$%s" expected value of type "%s" which cannot '
                    "be used as an input type."
                    % (name, print_ast(var_def.type)),
                    [var_def],
                )
            )
            continue

        if name not in variables:
            if var_def.default_value is not None:
                try:
                    coerced[name] = value_from_ast(
                        var_def.default_value, var_type
                    )
                except InvalidValue as err:
                    errors.append(
                        VariableCoercionError(
                            'Variable "$%s" got invalid default value %s (%s)'
                            % (name, print_ast(var_def.default_value), err),
                            [var_def],
                        )
                    )
            elif isinstance(var_type, NonNullType):
                errors.append(
                    VariableCoercionError(
                        'Variable "$%s" of required type "%s" was not provided.'
                        % (name, var_type),
                        [var_def],
                    )
                )
        else:
            value = variables[name]
            if value is None and isinstance(var_type, NonNullType):
                errors.append(
                    VariableCoercionError(
                        'Variable "$%s" of required type "%s" must not be null.'
                        % (name, var_type),
                        [var_def],
                    )
                )
            else:
                try:
                    coerced[name] = coerce_value(value, var_type)
                except MultiCoercionError as err:
                    for child_err in err.errors:
                        errors.append(
                            VariableCoercionError(
                                'Variable "$%s" got invalid value %s (%s)'
                                % (
                                    name,
                                    json.dumps(value, sort_keys=True),
                                    child_err,
                                ),
                                [var_def],
                            )
                        )
                except (InvalidValue, CoercionError) as err:
                    errors.append(
                        VariableCoercionError(
                            'Variable "$%s" got invalid value %s (%s)'
                            % (name, json.dumps(value, sort_keys=True), err),
                            [var_def],
                        )
                    )

    if errors:
        raise VariablesCoercionError(errors)

    return coerced



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import KnownDirectivesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_with_no_directives(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    query Foo {
        name
        ...Frag
    }

    fragment Frag on Dog {
        name
    }
    """,
        [],
        [],
    )


def test_with_known_directives(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    {
        dog @include(if: true) {
            name
        }
        human @skip(if: false) {
            name
        }
    }
    """,
        [],
        [],
    )


def test_with_unknown_directive(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    {
        dog @unknown(directive: "value") {
            name
        }
    }
    """,
        ['Unknown directive "@unknown"'],
        [(19, 47)],
    )


def test_with_many_unknown_directives(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    {
        dog @unknown(directive: "value") {
            name
        }
        human @unknown(directive: "value") {
            name
            pets @unknown(directive: "value") {
               name
            }
        }
    }
    """,
        [
            'Unknown directive "@unknown"',
            'Unknown directive "@unknown"',
            'Unknown directive "@unknown"',
        ],
        [(19, 47), (91, 119), (156, 184)],
    )


def test_with_well_placed_directives(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    query Foo @onQuery {
        name @include(if: true)
        ...Frag @include(if: true)
        skippedField @skip(if: true)
        ...SkippedFrag @skip(if: true)
    }

    mutation Bar @onMutation {
        someField
    }
    """,
        [],
        [],
    )


def test_with_misplaced_directives(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    query Foo @include(if: true) {
        name @onQuery
        ...Frag @onQuery
    }

    mutation Bar @onQuery {
        someField
    }
    """,
        [
            'Directive "@include" may not be used on QUERY',
            'Directive "@onQuery" may not be used on FIELD',
            'Directive "@onQuery" may not be used on FRAGMENT_SPREAD',
            'Directive "@onQuery" may not be used on MUTATION',
        ],
        [(15, 33), (49, 57), (74, 82), (107, 115)],
    )


def test_with_well_placed_directives_within_schema_language(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    type MyObj implements MyInterface @onObject {
        myField(myArg: Int @onArgumentDefinition): String @onFieldDefinition
    }

    extend type MyObj @onObject

    scalar MyScalar @onScalar

    extend scalar MyScalar @onScalar

    interface MyInterface @onInterface {
        myField(myArg: Int @onArgumentDefinition): String @onFieldDefinition
    }

    extend interface MyInterface @onInterface

    union MyUnion @onUnion = MyObj | Other

    extend union MyUnion @onUnion

    enum MyEnum @onEnum {
        MY_VALUE @onEnumValue
    }

    extend enum MyEnum @onEnum

    input MyInput @onInputObject {
        myField: Int @onInputFieldDefinition
    }

    extend input MyInput @onInputObject

    schema @onSchema {
        query: MyQuery
    }
    """,
        [],
        [],
    )


def test_with_misplaced_directives_within_schema_language(schema):
    run_test(
        KnownDirectivesChecker,
        schema,
        """
    type MyObj implements MyInterface @onInterface {
        myField(myArg: Int @onInputFieldDefinition): \
String @onInputFieldDefinition
    }

    scalar MyScalar @onEnum

    interface MyInterface @onObject {
        myField(myArg: Int @onInputFieldDefinition): \
String @onInputFieldDefinition
    }

    union MyUnion @onEnumValue = MyObj | Other

    enum MyEnum @onScalar {
        MY_VALUE @onUnion
    }

    input MyInput @onEnum {
        myField: Int @onArgumentDefinition
    }

    schema @onObject {
        query: MyQuery
    }
    """,
        [
            'Directive "@onInterface" may not be used on OBJECT',
            'Directive "@onInputFieldDefinition" may not be used on '
            "ARGUMENT_DEFINITION",
            'Directive "@onInputFieldDefinition" may not be used on '
            "FIELD_DEFINITION",
            'Directive "@onEnum" may not be used on SCALAR',
            'Directive "@onObject" may not be used on INTERFACE',
            'Directive "@onInputFieldDefinition" may not be used on '
            "ARGUMENT_DEFINITION",
            'Directive "@onInputFieldDefinition" may not be used on '
            "FIELD_DEFINITION",
            'Directive "@onEnumValue" may not be used on UNION',
            'Directive "@onScalar" may not be used on ENUM',
            'Directive "@onUnion" may not be used on ENUM_VALUE',
            'Directive "@onEnum" may not be used on INPUT_OBJECT',
            'Directive "@onArgumentDefinition" may not be used on '
            "INPUT_FIELD_DEFINITION",
            'Directive "@onObject" may not be used on SCHEMA',
        ],
    )



# -*- coding: utf-8 -*-

import re

import six

from ..lang import ast as _ast
from ..schema import (
    ID,
    EnumType,
    Float,
    InputObjectType,
    Int,
    ListType,
    NonNullType,
    ScalarType,
    is_input_type,
)
from ..schema.scalars import MAX_INT, MIN_INT, SPECIFIED_SCALAR_TYPES

_INT_RE = re.compile(r"^-?(0|[1-9][0-9]*)$")


def ast_node_from_value(value, input_type):  # noqa
    """ Infer an ast Node for a Python value given an input type.

    Args:
        value (any): Any python value that can be transformed into a node
        input_type (py_gql.schema.Type): Input type to consider

    Return:
        py_gql.lang.ast.Node: Inferred value node

    Raises:
        :py:class:`ValueError`: when coercion into a node fails
    """
    assert is_input_type(input_type), "Only supports input types"
    if isinstance(input_type, NonNullType):
        node = ast_node_from_value(value, input_type.type)
        if isinstance(node, _ast.NullValue):
            raise ValueError('Value of type "%s" cannot be null' % input_type)
        return node

    if value is None:
        return _ast.NullValue()

    if isinstance(input_type, ListType):
        if isinstance(value, (list, tuple)):
            return _ast.ListValue(
                values=[
                    ast_node_from_value(entry, input_type.type)
                    for entry in value
                ]
            )
        return ast_node_from_value(value, input_type.type)

    if isinstance(input_type, InputObjectType):
        if not isinstance(value, dict):
            raise ValueError('Value of type "%s" must be a dict' % input_type)

        field_nodes = []
        for field_def in input_type.fields:
            if field_def.name in value:
                field_value = ast_node_from_value(
                    value[field_def.name], field_def.type
                )
                field_nodes.append(
                    _ast.ObjectField(
                        name=_ast.Name(value=field_def.name), value=field_value
                    )
                )
            elif field_def.required:
                raise ValueError(
                    'Field "%s" of type "%s" is required'
                    % (field_def.name, input_type)
                )

        return _ast.ObjectValue(fields=field_nodes)

    if isinstance(input_type, ScalarType):
        serialized = input_type.serialize(value)
    elif isinstance(input_type, EnumType):
        serialized = input_type.get_name(value)

    if serialized is None:
        return _ast.NullValue()

    if isinstance(serialized, bool):
        return _ast.BooleanValue(value=serialized)

    if isinstance(serialized, (int, float)):
        if input_type is Int:
            return _ast.IntValue(value=str(serialized))
        elif input_type is Float:
            if int(serialized) == serialized and MIN_INT < serialized < MAX_INT:
                return _ast.IntValue(value=str(int(serialized)))
            elif MIN_INT < serialized < MAX_INT:
                return _ast.FloatValue(value=str(serialized))
        return _ast.FloatValue(value="%s" % serialized)

    if isinstance(serialized, six.string_types):
        if isinstance(input_type, EnumType):
            return _ast.EnumValue(value=serialized)
        elif input_type is ID and _INT_RE.match(serialized):
            return _ast.IntValue(value=serialized)
        elif (
            isinstance(input_type, ScalarType)
            and input_type not in SPECIFIED_SCALAR_TYPES
        ):
            if _INT_RE.match(serialized):
                intvalue = int(serialized)
                if MIN_INT < intvalue < MAX_INT:
                    return _ast.IntValue(value=serialized)
                else:
                    return _ast.FloatValue(value=serialized)
            try:
                fl = float(serialized)
            except ValueError:
                pass
            else:
                return _ast.FloatValue(value="%s" % fl)

        return _ast.StringValue(value=serialized)

    raise ValueError(
        'Cannot convert value %r of type "%s"' % (value, input_type)
    )



# -*- coding: utf-8 -*-

import pytest

from py_gql._string_utils import dedent
from py_gql.exc import SDLError
from py_gql.execution import execute
from py_gql.lang import parse
from py_gql.schema import UUID
from py_gql.schema.build import build_schema_from_ast, make_executable_schema
from py_gql.schema.directives import SPECIFIED_DIRECTIVES


def _check(schema):
    s = make_executable_schema(schema)
    assert s.to_string() == dedent(schema)
    return s


def test_built_schema_is_executable():
    schema = make_executable_schema(
        parse(
            """
            type Query {
                str: String
            }
            """
        )
    )
    data, _ = execute(
        schema, parse("{ str }"), initial_value={"str": 123}
    ).result()
    assert data == {"str": "123"}


def test_accepts_strings():
    schema = make_executable_schema(
        """
        type Query {
            str: String
        }
        """
    )
    data, _ = execute(
        schema, parse("{ str }"), initial_value={"str": 123}
    ).result()
    assert data == {"str": "123"}


def test_simple_type():
    schema = """
    type Query {
        str: String
        int: Int
        float: Float
        id: ID
        bool: Boolean
    }
    """
    _check(schema)


def test_with_directive():
    schema = """
    directive @foo(arg: Int) on FIELD

    type Query {
        str: String
    }
    """
    _check(schema)


def test_descriptions_supports():
    schema = '''
    """This is a directive"""
    directive @foo(
        """It has an argument"""
        arg: Int
    ) on FIELD

    """With an enum"""
    enum Color {
        RED

        """Not a creative color"""
        GREEN
        BLUE
    }

    """What a great type"""
    type Query {
        """And a field to boot"""
        str: String
    }
    '''
    _check(schema)


def test_specified_directives_are_enforced():
    schema = make_executable_schema(
        """
        directive @foo(arg: Int) on FIELD

        type Query {
            str: String
        }
        """
    )
    for d in SPECIFIED_DIRECTIVES:
        assert d is schema.directives[d.name]


def test_specified_directives_can_be_overriden():
    schema = make_executable_schema(
        """
        directive @skip on FIELD
        directive @include on FIELD
        directive @deprecated on FIELD_DEFINITION

        type Query {
            str: String
        }
        """
    )
    for d in SPECIFIED_DIRECTIVES:
        assert d is not schema.directives[d.name]


def test_type_modifiers():
    _check(
        """
        type Query {
            nonNullStr: String!
            listOfStrs: [String]
            listOfNonNullStrs: [String!]
            nonNullListOfStrs: [String]!
            nonNullListOfNonNullStrs: [String!]!
        }
        """
    )


def test_recursive_type():
    _check(
        """
        type Query {
            str: String
            recurse: Query
        }
        """
    )


def test_circular_types():
    _check(
        """
        schema {
            query: TypeOne
        }

        type TypeOne {
            str: String
            typeTwo: TypeTwo
        }

        type TypeTwo {
            str: String
            typeOne: TypeOne
        }
        """
    )


def test_single_argument_field():
    _check(
        """
        type Query {
            str(int: Int): String
            floatToStr(float: Float): String
            idToStr(id: ID): String
            booleanToStr(bool: Boolean): String
            strToStr(bool: String): String
        }
        """
    )


def test_multiple_arguments():
    _check(
        """
        type Query {
            str(int: Int, bool: Boolean): String
        }
        """
    )


def test_simple_interface():
    _check(
        """
        type Query implements WorldInterface {
            str: String
        }

        interface WorldInterface {
            str: String
        }
        """
    )


def test_simple_output_enum():
    _check(
        """
        enum Hello {
            WORLD
        }

        type Query {
            hello: Hello
        }
        """
    )


def test_simple_input_enum():
    _check(
        """
        enum Hello {
            WORLD
        }

        type Query {
            str(hello: Hello): String
        }
        """
    )


def test_multiple_values_enum():
    _check(
        """
        enum Hello {
            WO
            RLD
        }

        type Query {
            hello: Hello
        }
        """
    )


def test_union():
    _check(
        """
        union Hello = WorldOne | WorldTwo

        type Query {
            hello: Hello
        }

        type WorldOne {
            str: String
        }

        type WorldTwo {
            str: String
        }
        """
    )


def test_executing_union_default_resolve_type():
    schema = make_executable_schema(
        """
        type Query {
            fruits: [Fruit]
        }

        union Fruit = Apple | Banana

        type Apple {
            color: String
        }

        type Banana {
            length: Int
        }
        """
    )

    data, _ = execute(
        schema,
        parse(
            """
            {
                fruits {
                    ... on Apple {
                        color
                    }
                    ... on Banana {
                        length
                    }
                }
            }
            """
        ),
        initial_value={
            "fruits": [
                {"color": "green", "__typename__": "Apple"},
                {"length": 5, "__typename__": "Banana"},
            ]
        },
    ).result()

    assert data == {"fruits": [{"color": "green"}, {"length": 5}]}


def test_executing_interface_default_resolve_type():
    schema = make_executable_schema(
        """
        type Query {
            characters: [Character]
        }

        interface Character {
            name: String!
        }

        type Human implements Character {
            name: String!
            totalCredits: Int
        }

        type Droid implements Character {
            name: String!
            primaryFunction: String
        }
        """
    )

    data, _ = execute(
        schema,
        parse(
            """
            {
                characters {
                    name
                    ... on Human {
                        totalCredits
                    }
                    ... on Droid {
                        primaryFunction
                    }
                }
            }
            """
        ),
        initial_value={
            "characters": [
                {
                    "name": "Han Solo",
                    "totalCredits": 10,
                    "__typename__": "Human",
                },
                {
                    "name": "R2-D2",
                    "primaryFunction": "Astromech",
                    "__typename__": "Droid",
                },
            ]
        },
    ).result()

    assert data == {
        "characters": [
            {"name": "Han Solo", "totalCredits": 10},
            {"name": "R2-D2", "primaryFunction": "Astromech"},
        ]
    }


def test_custom_scalar():
    schema = _check(
        """
        scalar CustomScalar

        type Query {
            customScalar: CustomScalar
        }
        """
    )
    scalar = schema.types["CustomScalar"]
    assert scalar.parse("foo") == "foo"
    assert scalar.serialize(123) == 123


def test_input_object():
    _check(
        """
        input Input {
            int: Int
        }

        type Query {
            field(in: Input): String
        }
        """
    )


def test_input_object_with_default_value():
    _check(
        """
        input Input {
            int: Int = 42
        }

        type Query {
            field(in: Input): String
        }
        """
    )


def test_simple_argument_field_with_default():
    _check(
        """
        type Query {
            str(int: Int = 2): String
        }
        """
    )


def test_custom_scalar_argument_field_with_default():
    _check(
        """
        scalar CustomScalar

        type Query {
            str(int: CustomScalar = 2): String
        }
        """
    )


def test_simple_type_with_mutation():
    _check(
        """
        schema {
            query: HelloScalars
            mutation: Mutation
        }

        type HelloScalars {
            str: String
            int: Int
            bool: Boolean
        }

        type Mutation {
            addHelloScalars(str: String, int: Int, bool: Boolean): HelloScalars
        }
        """
    )


def test_simple_type_with_subscription():
    _check(
        """
        schema {
            query: HelloScalars
            subscription: Subscription
        }

        type HelloScalars {
            str: String
            int: Int
            bool: Boolean
        }

        type Subscription {
            sbscribeHelloScalars(str: String, int: Int, bool: Boolean): HelloScalars
        }
        """
    )


def test_unreferenced_type_implementing_referenced_interface():
    _check(
        """
        type Concrete implements Iface {
            key: String
        }

        interface Iface {
            key: String
        }

        type Query {
            iface: Iface
        }
        """
    )


def test_unreferenced_type_implementing_referenced_union():
    _check(
        """
        type Concrete {
            key: String
        }

        type Query {
            union: Union
        }

        union Union = Concrete
        """
    )


def test_supports_deprecated():
    _check(
        """
        enum MyEnum {
            VALUE
            OLD_VALUE @deprecated
            OTHER_VALUE @deprecated(reason: "Terrible reasons")
        }

        type Query {
            field1: String @deprecated
            field2: Int @deprecated(reason: "Because I said so")
            enum: MyEnum
        }
        """
    )


def test_root_operation_types_with_custom_names():
    schema = make_executable_schema(
        """
        schema {
            query: SomeQuery
            mutation: SomeMutation
            subscription: SomeSubscription
        }
        type SomeQuery { str: String }
        type SomeMutation { str: String }
        type SomeSubscription { str: String }
        """
    )

    assert schema.query_type.name == "SomeQuery"
    assert schema.mutation_type.name == "SomeMutation"
    assert schema.subscription_type.name == "SomeSubscription"


def test_default_root_operation_type_names():
    schema = make_executable_schema(
        """
        type Query { str: String }
        type Mutation { str: String }
        type Subscription { str: String }
        """
    )

    assert schema.query_type.name == "Query"
    assert schema.mutation_type.name == "Mutation"
    assert schema.subscription_type.name == "Subscription"


def test_allows_only_a_single_schema_definition():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
            }

            schema {
                query: Hello
            }

            type Hello {
                bar: Bar
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 6}],
        "message": "Must provide only one schema definition",
    }


def test_allows_only_a_single_query_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
                query: Yellow
            }

            type Hello {
                bar: String
            }

            type Yellow {
                isColor: Boolean
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 2}, {"column": 17, "line": 4}],
        "message": "Schema must only define a single query operation",
    }


def test_allows_only_a_single_mutation_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
                mutation: Hello
                mutation: Yellow
            }

            type Hello {
                bar: String
            }

            type Yellow {
                isColor: Boolean
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 2}, {"column": 17, "line": 5}],
        "message": "Schema must only define a single mutation operation",
    }


def test_allows_only_a_single_subscription_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
                subscription: Hello
                subscription: Yellow
            }

            type Hello {
                bar: String
            }

            type Yellow {
                isColor: Boolean
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 2}, {"column": 17, "line": 5}],
        "message": "Schema must only define a single subscription operation",
    }


def test_unknown_type_referenced():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
            }

            type Hello {
                bar: Bar
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 22, "line": 7}],
        "message": "Type Bar not found in document",
    }


def test_unknown_type_in_interface_list():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema("type Query implements Bar { field: String }")
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 23, "line": 1}],
        "message": "Type Bar not found in document",
    }


def test_unknown_type_in_union_list():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            union TestUnion = Bar
            type Query { testUnion: TestUnion }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 31, "line": 2}],
        "message": "Type Bar not found in document",
    }


def test_unknown_query_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Wat
            }

            type Hello {
                str: String
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 24, "line": 3}],
        "message": "Type Wat not found in document",
    }


def test_unknown_mutation_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
                mutation: Wat
            }

            type Hello {
                str: String
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 27, "line": 4}],
        "message": "Type Wat not found in document",
    }


def test_unknown_subscription_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Hello
                subscription: Wat
            }

            type Hello {
                str: String
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 31, "line": 4}],
        "message": "Type Wat not found in document",
    }


def test_does_not_consider_operation_names_or_fragment_name():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Foo
            }

            query Foo { field }

            fragment Foo on Type { field }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 24, "line": 3}],
        "message": "Type Foo not found in document",
    }


def test_forbids_duplicate_type_definitions():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            schema {
                query: Repeated
            }

            type Repeated {
                id: Int
            }

            type Repeated {
                id: String
            }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 10}],
        "message": "Duplicate type Repeated",
    }


def test_forbids_duplicate_directive_definition():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: String
            }

            directive @foo on FIELD
            directive @foo on MUTATION
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 7}],
        "message": "Duplicate directive @foo",
    }


def test_inject_custom_types():
    schema = make_executable_schema(
        """
        type Query {
            foo: UUID
        }
        """,
        additional_types=[UUID],
    )
    assert schema.types["UUID"] is UUID


def test_inject_resolvers():
    resolvers = {"Query": {"foo": lambda *_: "foo"}}

    schema = make_executable_schema(
        """
        type Query {
            foo: String
        }
        """,
        resolvers=resolvers,
    )

    assert schema.query_type.fields[0].resolve() == "foo"


def test_inject_resolvers_as_flat_map():
    resolvers = {"Query.foo": lambda *_: "foo"}

    schema = make_executable_schema(
        """
        type Query {
            foo: String
        }
        """,
        resolvers=resolvers,
    )

    assert schema.query_type.fields[0].resolve() == "foo"


def test_inject_resolvers_as_callable():
    resolvers = {"Query.foo": lambda *_: "foo"}

    schema = make_executable_schema(
        """
        type Query {
            foo: String
        }
        """,
        resolvers=lambda t, f: resolvers.get(t + "." + f),
    )

    assert schema.query_type.fields[0].resolve() == "foo"


def test_build_schema_from_ast_ignores_extensions():
    assert (
        build_schema_from_ast(
            """
            type Query {
                one: Int
            }

            extend type Object {
                one: Int
            }

            extend type Query {
                two: String
            }
            """
        ).to_string()
        == dedent(
            """
            type Query {
                one: Int
            }
            """
        )
    )



# -*- coding: utf-8 -*-
""" """

from itertools import chain

from . import rules as _rules
from ..lang.visitor import ParrallelVisitor, visit
from ..utilities import TypeInfoVisitor
from .visitors import ValidationVisitor  # noqa: F401

SPECIFIED_RULES = (
    _rules.ExecutableDefinitionsChecker,
    _rules.UniqueOperationNameChecker,
    _rules.LoneAnonymousOperationChecker,
    _rules.SingleFieldSubscriptionsChecker,
    _rules.KnownTypeNamesChecker,
    _rules.FragmentsOnCompositeTypesChecker,
    _rules.VariablesAreInputTypesChecker,
    _rules.ScalarLeafsChecker,
    _rules.FieldsOnCorrectTypeChecker,
    _rules.UniqueFragmentNamesChecker,
    _rules.KnownFragmentNamesChecker,
    _rules.NoUnusedFragmentsChecker,
    _rules.PossibleFragmentSpreadsChecker,
    _rules.NoFragmentCyclesChecker,
    _rules.UniqueVariableNamesChecker,
    _rules.NoUndefinedVariablesChecker,
    _rules.NoUnusedVariablesChecker,
    _rules.KnownDirectivesChecker,
    _rules.UniqueDirectivesPerLocationChecker,
    _rules.KnownArgumentNamesChecker,
    _rules.UniqueArgumentNamesChecker,
    _rules.ValuesOfCorrectTypeChecker,
    _rules.ProvidedRequiredArgumentsChecker,
    _rules.VariablesInAllowedPositionChecker,
    _rules.OverlappingFieldsCanBeMergedChecker,
    _rules.UniqueInputFieldNamesChecker,
)


class ValidationResult(object):
    """ Encode validation result by wrapping a collection of
    :class:`~py_gql.exc.ValidationError`.

    Instances are falsy when they contain at least one error.
    """

    def __init__(self, errors):
        #: errors: List[:class:`~py_gql.exc.ValidationError`]
        self.errors = errors or []

    def __bool__(self):
        return not self.errors

    __nonzero__ = __bool__


def validate_ast(schema, ast_root, validators=None):
    """ Check that an ast is a valid GraphQL query document.

    Runs a parse tree through a list of
    :class:`~py_gql.validation.ValidationVisitor` given a
    :class:`~py_gql.schema.Schema` instance.

    Warning:
        This assumes the ast is a valid document generated by
        :func:`py_gql.lang.parse` and will most likely break
        unexpectedly if that's not the case.

    Args:
        schema (py_gql.schema.Schema):
            Schema to validate against (for known types, directives, etc.).

        ast_root (py_gql.lang.ast.Document): The parse tree root.

        validators (list):
            List of validator classes (subclasses of
            :class:`py_gql.validation.ValidationVisitor`) to use. Can also
            contain tuples of the shape ``(cls, kw)`` where ``kw`` will be
            passed to ``cls`` on instantiation.
            Defaults to :obj:`~py_gql.validation.SPECIFIED_RULES`.

    Returns:
        py_gql.validation.ValidationResult: Validation result wrapping any
        validatione error that occured.
    """
    type_info = TypeInfoVisitor(schema)
    if validators is None:
        validators = SPECIFIED_RULES

    def instantiate_validator(cls_or_tuple, schema, type_info):
        if isinstance(cls_or_tuple, tuple):
            cls, kw = cls_or_tuple
        else:
            cls, kw = cls_or_tuple, {}
        assert issubclass(cls, ValidationVisitor)
        return cls(schema, type_info, **kw)

    # Type info NEEDS to be first to be accurately used inside other validators
    # so when a validator enters node the type stack has already been updated.
    validator = ParrallelVisitor(
        type_info,
        *[
            instantiate_validator(validator_, schema, type_info)
            for validator_ in validators
        ]
    )

    visit(validator, ast_root)
    return ValidationResult(
        list(chain(*[v.errors for v in validator.visitors[1:]]))
    )



# -*- coding: utf-8 -*-
"""
"""

from py_gql.lang import ast as _ast
from py_gql.lang.parser import parse
from py_gql.lang.visitor import SkipNode, Visitor, visit


class NullVisitor(Visitor):
    def enter(self, node):
        pass

    def leave(self, node):
        pass


def test_null_visitor_does_not_crash():
    ast = parse("{ a }", no_location=True)
    visit(NullVisitor(), ast)


def test_null_visitor_does_not_crash_on_kitchen_sink(fixture_file):
    source = fixture_file("kitchen-sink.graphql")
    ast = parse(source, no_location=True)
    visit(NullVisitor(), ast)


def test_null_visitor_does_not_crash_on_kitchen_sink_schema(fixture_file):
    source = fixture_file("schema-kitchen-sink.graphql")
    ast = parse(source, no_location=True)
    visit(NullVisitor(), ast)


class Tracker(NullVisitor):
    def __init__(self):
        self.stack = []

    def enter(self, node):
        self.stack.append(("enter", node.__class__.__name__))

    def leave(self, node):
        self.stack.append(("leave", node.__class__.__name__))


def test_it_processes_nodes_in_the_correct_order():
    ast = parse("{ a }", no_location=True)
    visitor = Tracker()
    visit(visitor, ast)
    assert visitor.stack == [
        ("enter", "Document"),
        ("enter", "OperationDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("leave", "Document"),
    ]


def test_it_allows_early_exit():
    """ interrupt processing by raising """
    ast = parse("{ a { b { c { d { e } } } } }", no_location=True)

    class _Visitor(Tracker):
        def enter(self, node):
            super(_Visitor, self).enter(node)
            if isinstance(node, _ast.Field) and node.name.value == "b":
                raise SkipNode()

    visitor = _Visitor()
    visit(visitor, ast)

    assert visitor.stack == [
        ("enter", "Document"),
        ("enter", "OperationDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("leave", "Document"),
    ]


def test_it_processes_kitchen_sink(fixture_file):
    ks = fixture_file("kitchen-sink.graphql")
    visitor = Tracker()
    visit(visitor, parse(ks, no_location=True))
    assert visitor.stack == [
        ("enter", "Document"),
        ("enter", "OperationDefinition"),
        ("enter", "VariableDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "VariableDefinition"),
        ("enter", "VariableDefinition"),
        ("enter", "EnumValue"),
        ("leave", "EnumValue"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "VariableDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "ListValue"),
        ("enter", "IntValue"),
        ("leave", "IntValue"),
        ("enter", "IntValue"),
        ("leave", "IntValue"),
        ("leave", "ListValue"),
        ("leave", "Argument"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("enter", "InlineFragment"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "IntValue"),
        ("leave", "IntValue"),
        ("leave", "Argument"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("enter", "Directive"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("leave", "Directive"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("enter", "FragmentSpread"),
        ("leave", "FragmentSpread"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "InlineFragment"),
        ("enter", "InlineFragment"),
        ("enter", "Directive"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("leave", "Directive"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "InlineFragment"),
        ("enter", "InlineFragment"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "InlineFragment"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("enter", "OperationDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "IntValue"),
        ("leave", "IntValue"),
        ("leave", "Argument"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("enter", "OperationDefinition"),
        ("enter", "VariableDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "VariableDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("enter", "Field"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("enter", "FragmentDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("enter", "Argument"),
        ("enter", "Variable"),
        ("leave", "Variable"),
        ("leave", "Argument"),
        ("enter", "Argument"),
        ("enter", "ObjectValue"),
        ("enter", "ObjectField"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "ObjectField"),
        ("enter", "ObjectField"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "ObjectField"),
        ("leave", "ObjectValue"),
        ("leave", "Argument"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "FragmentDefinition"),
        ("enter", "OperationDefinition"),
        ("enter", "SelectionSet"),
        ("enter", "Field"),
        ("enter", "Argument"),
        ("enter", "BooleanValue"),
        ("leave", "BooleanValue"),
        ("leave", "Argument"),
        ("enter", "Argument"),
        ("enter", "BooleanValue"),
        ("leave", "BooleanValue"),
        ("leave", "Argument"),
        ("enter", "Argument"),
        ("enter", "NullValue"),
        ("leave", "NullValue"),
        ("leave", "Argument"),
        ("leave", "Field"),
        ("enter", "Field"),
        ("leave", "Field"),
        ("leave", "SelectionSet"),
        ("leave", "OperationDefinition"),
        ("leave", "Document"),
    ]


def test_it_processes_schema_kitchen_sink(fixture_file):
    ks = fixture_file("schema-kitchen-sink.graphql")
    visitor = Tracker()
    visit(visitor, parse(ks, no_location=True))
    assert visitor.stack == [
        ("enter", "Document"),
        ("enter", "SchemaDefinition"),
        ("enter", "OperationTypeDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "OperationTypeDefinition"),
        ("enter", "OperationTypeDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "OperationTypeDefinition"),
        ("leave", "SchemaDefinition"),
        ("enter", "ObjectTypeDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "InputValueDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "ListType"),
        ("leave", "ListType"),
        ("enter", "ListValue"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "ListValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "ObjectValue"),
        ("enter", "ObjectField"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "ObjectField"),
        ("leave", "ObjectValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NullValue"),
        ("leave", "NullValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("leave", "ObjectTypeDefinition"),
        ("enter", "ObjectTypeDefinition"),
        ("enter", "Directive"),
        ("enter", "Argument"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "Argument"),
        ("leave", "Directive"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "InputValueDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "FieldDefinition"),
        ("leave", "ObjectTypeDefinition"),
        ("enter", "ObjectTypeDefinition"),
        ("leave", "ObjectTypeDefinition"),
        ("enter", "ObjectTypeExtension"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "ListType"),
        ("leave", "ListType"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("leave", "ObjectTypeExtension"),
        ("enter", "ObjectTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "ObjectTypeExtension"),
        ("enter", "InterfaceTypeDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "FieldDefinition"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "StringValue"),
        ("leave", "StringValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("leave", "InterfaceTypeDefinition"),
        ("enter", "InterfaceTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "InputValueDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "FieldDefinition"),
        ("leave", "InterfaceTypeDefinition"),
        ("enter", "InterfaceTypeDefinition"),
        ("leave", "InterfaceTypeDefinition"),
        ("enter", "InterfaceTypeExtension"),
        ("enter", "FieldDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("leave", "FieldDefinition"),
        ("leave", "InterfaceTypeExtension"),
        ("enter", "InterfaceTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "InterfaceTypeExtension"),
        ("enter", "UnionTypeDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "UnionTypeDefinition"),
        ("enter", "UnionTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "UnionTypeDefinition"),
        ("enter", "UnionTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "UnionTypeDefinition"),
        ("enter", "UnionTypeDefinition"),
        ("leave", "UnionTypeDefinition"),
        ("enter", "UnionTypeExtension"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "UnionTypeExtension"),
        ("enter", "UnionTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "UnionTypeExtension"),
        ("enter", "ScalarTypeDefinition"),
        ("leave", "ScalarTypeDefinition"),
        ("enter", "ScalarTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "ScalarTypeDefinition"),
        ("enter", "ScalarTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "ScalarTypeExtension"),
        ("enter", "EnumTypeDefinition"),
        ("enter", "EnumValueDefinition"),
        ("leave", "EnumValueDefinition"),
        ("enter", "EnumValueDefinition"),
        ("leave", "EnumValueDefinition"),
        ("leave", "EnumTypeDefinition"),
        ("enter", "EnumTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "EnumValueDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "EnumValueDefinition"),
        ("enter", "EnumValueDefinition"),
        ("leave", "EnumValueDefinition"),
        ("leave", "EnumTypeDefinition"),
        ("enter", "EnumTypeDefinition"),
        ("leave", "EnumTypeDefinition"),
        ("enter", "EnumTypeExtension"),
        ("enter", "EnumValueDefinition"),
        ("leave", "EnumValueDefinition"),
        ("leave", "EnumTypeExtension"),
        ("enter", "EnumTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "EnumTypeExtension"),
        ("enter", "InputObjectTypeDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "IntValue"),
        ("leave", "IntValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "InputObjectTypeDefinition"),
        ("enter", "InputObjectTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "InputValueDefinition"),
        ("leave", "InputObjectTypeDefinition"),
        ("enter", "InputObjectTypeDefinition"),
        ("leave", "InputObjectTypeDefinition"),
        ("enter", "InputObjectTypeExtension"),
        ("enter", "InputValueDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("enter", "FloatValue"),
        ("leave", "FloatValue"),
        ("leave", "InputValueDefinition"),
        ("leave", "InputObjectTypeExtension"),
        ("enter", "InputObjectTypeExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "InputObjectTypeExtension"),
        ("enter", "DirectiveDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("leave", "DirectiveDefinition"),
        ("enter", "DirectiveDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("leave", "DirectiveDefinition"),
        ("enter", "DirectiveDefinition"),
        ("enter", "InputValueDefinition"),
        ("enter", "NonNullType"),
        ("leave", "NonNullType"),
        ("leave", "InputValueDefinition"),
        ("leave", "DirectiveDefinition"),
        ("enter", "SchemaExtension"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "SchemaExtension"),
        ("enter", "SchemaExtension"),
        ("enter", "OperationTypeDefinition"),
        ("enter", "NamedType"),
        ("leave", "NamedType"),
        ("leave", "OperationTypeDefinition"),
        ("enter", "Directive"),
        ("leave", "Directive"),
        ("leave", "SchemaExtension"),
        ("leave", "Document"),
    ]



# -*- coding: utf-8 -*-
""" Middlewares provide support for modifying / short-circuiting the field
resolution part of the execution. A middleware is a callable that returns /
yields to the next step.

.. highlight:: python

    # Functional middleware
    def ensure_bar_middleware(next_, root, args, context, info):
        if args.get('foo') != 'bar':
            raise ResolverError('Not bar!')
        return next_(root, args, context, info)

    # Generator middleware
    def logging_middleware(next_, root, args, context, info):
        logger.debug('start', info.path)
        yield next_(root, args, context, info)
        logger.debug('end', info.path)

    # Class based middleware
    class CollectFieldsMiddleware(object):
        def __init__(self):
            self.fields = []

        def __call__(self, next_, root, args, context, info):
            self.fields.append(info.path)
            return next_(root, args, context, info)

    collect_fields = CollectFieldsMiddleware()

"""

import functools as ft
from inspect import isgeneratorfunction

from . import _concurrency


def _is_generator(mw):
    return isgeneratorfunction(mw) or isgeneratorfunction(mw.__call__)


def apply_middlewares(func, middlewares):
    """ Apply middleware functions to a base function.

    - Middlewares must the signature: ``(next, *args, **kwargs) -> any`` where
      ``(*args, **kwargs) -> any`` is the signature of the wrapped function.
    - They can either ``return`` or ``yield`` in order to have clean up logic
    - Generator based middlewares **must** yield at least once
    - Middlewares are evaluated inside-out
    """
    if not middlewares:
        return func

    tail = func
    for mw in reversed(middlewares):
        assert callable(mw)
        if _is_generator(mw):
            mw = generator_middleware(mw)
        tail = ft.partial(mw, tail)

    return tail


def generator_middleware(func):
    """ Transform a middleware defined using the ``yield`` keyword into
    a usable middleware function.

    Middlewares defined as such are expected to ``yield`` only once and any
    non-yielding code present after the ``yield`` keyword is guaranteed to be
    run. """

    def wrapped(step, *args, **kwargs):
        gen = func(step, *args, **kwargs)

        def _finish(_):
            try:
                next(gen)
            except StopIteration:
                pass

        try:
            res = next(gen)
        except StopIteration:
            raise RuntimeError("Generator middleware did not yield")

        if _concurrency.is_deferred(res):
            res.add_done_callback(_finish)
        else:
            _finish(None)
        return res

    return wrapped



# -*- coding: utf-8 -*-
""" Export schema as SDL. """
from __future__ import unicode_literals

import operator as op

from .._string_utils import leading_whitespace, wrapped_lines
from ..lang import print_ast
from ..utilities import ast_node_from_value
from .directives import DEFAULT_DEPRECATION, SPECIFIED_DIRECTIVES
from .introspection import is_introspection_type
from .scalars import SPECIFIED_SCALAR_TYPES, String
from .types import (
    EnumType,
    InputObjectType,
    InterfaceType,
    ObjectType,
    ScalarType,
    UnionType,
)


class SchemaPrinter(object):
    """
    Args:
        indent (Union[str, int]): Indent character or number of spaces

        include_descriptions (bool):
            If ``True`` include descriptions in the output

        description_format ("comments"|"block"):
            Control how descriptions are formatted. ``"comments"`` is the
            old standard and will be compatible with most GraphQL parsers
            while ``"block"`` is part of the most recent specification and
            includes descriptions as block strings that can be extracted
            according to the specification.

        include_introspection (bool):
            If ``True``, include introspection types in the output
    """

    __slots__ = (
        "indent",
        "include_descriptions",
        "description_format",
        "include_introspection",
    )

    def __init__(
        self,
        indent=4,
        include_descriptions=True,
        description_format="block",
        include_introspection=False,
    ):
        self.include_descriptions = include_descriptions
        self.description_format = description_format
        if isinstance(indent, int):
            self.indent = indent * " "
        else:
            self.indent = indent

        self.include_introspection = include_introspection

    def __call__(self, schema):
        """
        schema (py_gql.schema.Schema): Schema to format

        Returns:
            str: Formatted GraphQL schema
        """
        directives = sorted(
            [
                d
                for d in schema.directives.values()
                if d not in SPECIFIED_DIRECTIVES
            ],
            key=op.attrgetter("name"),
        )

        types = sorted(
            [
                t
                for t in schema.types.values()
                if (
                    t not in SPECIFIED_SCALAR_TYPES
                    and (
                        self.include_introspection
                        or not is_introspection_type(t)
                    )
                )
            ],
            key=op.attrgetter("name"),
        )

        parts = (
            [_schema_definition(schema, self.indent)]
            + [
                self.print_directive(d)
                for d in (
                    SPECIFIED_DIRECTIVES if self.include_introspection else []
                )
            ]
            + [self.print_directive(d) for d in directives]
            + [self.print_type(t) for t in types]
        )

        return "\n\n".join((p for p in parts if p)) + "\n"

    def print_description(self, definition, depth=0, first_in_block=True):
        """ Format an object description according to current configuration.

        Args:
            definitions (any): Described object
            depth (int): Level of indentation
            first_in_block (bool):

        Returns:
            str:
        """
        if not self.include_descriptions or not getattr(
            definition, "description"
        ):
            return ""

        indent = self.indent * depth

        if self.description_format == "comments":
            prefix = indent + "# "
            max_len = 120 - len(prefix)
            lines = [
                (prefix + line.rstrip()) if line else "#"
                for line in wrapped_lines(
                    definition.description.split("\n"), max_len
                )
            ]
            return (
                ("\n" if not first_in_block else "") + "\n".join(lines) + "\n"
            )

        elif self.description_format == "block":
            max_len = 120 - len(indent)
            lines = list(
                wrapped_lines(definition.description.split("\n"), max_len)
            )
            first = lines[0]

            if len(lines) == 1 and len(first) < 70 and not first.endswith('"'):
                body = first.replace('"""', '\\"""')
            else:
                has_leading_whitespace = leading_whitespace(first)
                body = (
                    "\n".join(
                        [
                            "%s%s%s"
                            % (
                                "\n"
                                if (i == 0 and not has_leading_whitespace)
                                else "",
                                indent
                                if (i > 0 or not has_leading_whitespace)
                                else "",
                                line.replace('"""', '\\"""'),
                            )
                            for i, line in enumerate(lines)
                        ]
                    )
                    + "\n"
                    + indent
                )

            return '%s%s"""%s"""\n' % (
                "\n" if indent and not first_in_block else "",
                indent,
                body,
            )

        raise ValueError(
            "Invalid description format %s" % self.description_format
        )

    def print_deprecated(self, field_or_enum_value):
        if not field_or_enum_value.deprecated:
            return ""
        elif (
            not field_or_enum_value.deprecation_reason
            or field_or_enum_value.deprecation_reason == DEFAULT_DEPRECATION
        ):
            return " @deprecated"
        return " @deprecated(reason: %s)" % print_ast(
            ast_node_from_value(field_or_enum_value.deprecation_reason, String)
        )

    def print_type(self, type_):
        if isinstance(type_, ScalarType):
            return self.print_scalar_type(type_)
        elif isinstance(type_, ObjectType):
            return self.print_object_type(type_)
        elif isinstance(type_, InterfaceType):
            return self.print_interface_type(type_)
        elif isinstance(type_, UnionType):
            return self.print_union_type(type_)
        elif isinstance(type_, EnumType):
            return self.print_enum_type(type_)
        elif isinstance(type_, InputObjectType):
            return self.print_input_object_type(type_)

        raise TypeError(type_)

    def print_scalar_type(self, type_):
        """
        """
        return "%sscalar %s" % (
            self.print_description(type_, 0, True),
            type_.name,
        )

    def print_enum_type(self, type_):
        return "%senum %s {\n%s\n}" % (
            self.print_description(type_),
            type_.name,
            "\n".join(
                [
                    "".join(
                        [
                            self.print_description(enum_value, 1, not i),
                            self.indent,
                            enum_value.name,
                            self.print_deprecated(enum_value),
                        ]
                    )
                    for i, enum_value in enumerate(type_.values)
                ]
            ),
        )

    def print_union_type(self, type_):
        return "%sunion %s = %s" % (
            self.print_description(type_),
            type_.name,
            " | ".join((t.name for t in type_.types)),
        )

    def print_object_type(self, type_):
        return "%stype %s%s {\n%s\n}" % (
            self.print_description(type_),
            type_.name,
            " implements %s"
            % " & ".join([iface.name for iface in type_.interfaces])
            if type_.interfaces
            else "",
            self.print_fields(type_),
        )

    def print_interface_type(self, type_):
        return "%sinterface %s {\n%s\n}" % (
            self.print_description(type_),
            type_.name,
            self.print_fields(type_),
        )

    def print_fields(self, type_):
        return "\n".join(
            [
                "".join(
                    [
                        self.print_description(field, 1, not i),
                        self.indent,
                        field.name,
                        self.print_arguments(field.args, 1),
                        ": %s" % field.type,
                        self.print_deprecated(field),
                    ]
                )
                for i, field in enumerate(type_.fields)
            ]
        )

    def print_input_object_type(self, type_):
        return "%sinput %s {\n%s\n}" % (
            self.print_description(type_),
            type_.name,
            "\n".join(
                [
                    "%s%s%s"
                    % (
                        self.print_description(field, 1, not i),
                        self.indent,
                        self.print_input_value(field),
                    )
                    for i, field in enumerate(type_.fields)
                ]
            ),
        )

    def print_directive(self, directive):
        return "%sdirective @%s%s on %s" % (
            self.print_description(directive),
            directive.name,
            self.print_arguments(directive.arguments, 0),
            " | ".join(directive.locations),
        )

    def print_arguments(self, args, depth=0):
        if not args:
            return ""

        indent = self.indent * depth
        if self.include_descriptions and any(a.description for a in args):
            return "%s(\n%s\n%s)" % (
                indent,
                "\n".join(
                    [
                        "".join(
                            [
                                self.print_description(arg, depth + 1, not i),
                                self.indent + indent,
                                self.print_input_value(arg),
                            ]
                        )
                        for i, arg in enumerate(args)
                    ]
                ),
                indent,
            )
        else:
            return "(%s)" % ", ".join(
                [self.print_input_value(arg) for arg in args]
            )

    def print_input_value(self, arg_or_inut_field):
        s = "%s: %s" % (arg_or_inut_field.name, arg_or_inut_field.type)
        if arg_or_inut_field.has_default_value:
            s += " = %s" % print_ast(
                ast_node_from_value(
                    arg_or_inut_field.default_value, arg_or_inut_field.type
                )
            )
        return s


def _schema_definition(schema, indent):
    if (
        (not schema.query_type or schema.query_type.name == "Query")
        and (
            not schema.mutation_type or schema.mutation_type.name == "Mutation"
        )
        and (
            not schema.subscription_type
            or schema.subscription_type.name == "Subscription"
        )
    ):
        return ""

    operation_types = []
    if schema.query_type:
        operation_types.append("%squery: %s" % (indent, schema.query_type.name))
    if schema.mutation_type:
        operation_types.append(
            "%smutation: %s" % (indent, schema.mutation_type.name)
        )
    if schema.subscription_type:
        operation_types.append(
            "%ssubscription: %s" % (indent, schema.subscription_type.name)
        )

    return "schema {\n%s\n}" % ("\n".join(operation_types))



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import UniqueOperationNameChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_no_operations(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """fragment fragA on Type {
        field
    }""",
    )


def test_one_anon_operation(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """{
        field
    }""",
    )


def test_one_named_operation(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """query {
        field
    }""",
    )


def test_multiple_operations(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        field
    }

    query Bar {
        field
    }
    """,
    )


def test_multiple_operations_of_different_types(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        field
    }

    mutation Bar {
        field
    }

    subscription Baz {
        field
    }
    """,
    )


def test_fragment_and_operation_named_the_same(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        ...Foo
    }
    fragment Foo on Type {
        field
    }
    """,
    )


def test_multiple_operations_of_same_name(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        fieldA
    }

    query Foo {
        fieldB
    }
    """,
        ['Duplicate operation "Foo".'],
    )


def test_multiple_ops_of_same_name_of_different_types_mutation(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        fieldA
    }

    mutation Foo {
        fieldB
    }
    """,
        ['Duplicate operation "Foo".'],
    )


def test_multiple_ops_of_same_name_of_different_types_subscription(schema):
    run_test(
        UniqueOperationNameChecker,
        schema,
        """
    query Foo {
        fieldA
    }

    subscription Foo {
        fieldB
    }
    """,
        ['Duplicate operation "Foo".'],
    )



# -*- coding: utf-8 -*-
""" Global fixtures """

import os

import pytest


@pytest.fixture
def fixture_file():
    """ Helper to load fixture files by name. """

    def load(name):
        filepath = os.path.join(os.path.dirname(__file__), "_fixtures", name)
        with open(filepath, "rb") as f:
            return f.read().decode("utf-8")

    return load


@pytest.fixture
def starwars_schema():
    from ._star_wars import StarWarsSchema

    return StarWarsSchema


@pytest.fixture
def raiser():
    def factory(cls, *args, **kwargs):
        assert issubclass(cls, Exception)

        def _raiser(*a, **kw):
            raise cls(*args, **kwargs)

        return _raiser

    return factory



# -*- coding: utf-8 -*-

import pytest

from py_gql.execution import execute
from py_gql.lang import parse


def test_it_correctly_identifies_r2_d2_as_the_hero_of_the_star_wars_saga(
    starwars_schema
):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query HeroNameQuery {
                hero {
                name
                }
            }
            """
        ),
    ).result()
    assert result == {"hero": {"name": "R2-D2"}}
    assert errors == []


def test_id_and_friends_of_r2_d2(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query HeroNameAndFriendsQuery {
                hero {
                    id
                    name
                    friends {
                        name
                    }
                }
            }
            """
        ),
    ).result()
    assert result == {
        "hero": {
            "id": "2001",
            "name": "R2-D2",
            "friends": [
                {"name": "Luke Skywalker"},
                {"name": "Han Solo"},
                {"name": "Leia Organa"},
            ],
        }
    }
    assert errors == []


def test_the_friends_of_friends_of_r2_d2(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query NestedQuery {
                hero {
                    name
                    friends {
                        name
                        appearsIn
                        friends {
                        name
                        }
                    }
                }
            }
            """
        ),
    ).result()
    assert result == {
        "hero": {
            "name": "R2-D2",
            "friends": [
                {
                    "name": "Luke Skywalker",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Han Solo"},
                        {"name": "Leia Organa"},
                        {"name": "C-3PO"},
                        {"name": "R2-D2"},
                    ],
                },
                {
                    "name": "Han Solo",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Luke Skywalker"},
                        {"name": "Leia Organa"},
                        {"name": "R2-D2"},
                    ],
                },
                {
                    "name": "Leia Organa",
                    "appearsIn": ["NEWHOPE", "EMPIRE", "JEDI"],
                    "friends": [
                        {"name": "Luke Skywalker"},
                        {"name": "Han Solo"},
                        {"name": "C-3PO"},
                        {"name": "R2-D2"},
                    ],
                },
            ],
        }
    }
    assert errors == []


def test_luke_skywalker_using_id(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query FetchLukeQuery {
                human(id: "1000") {
                name
                }
            }
            """
        ),
    ).result()
    assert result == {"human": {"name": "Luke Skywalker"}}
    assert errors == []


@pytest.mark.parametrize(
    "id_, expected",
    [
        ("1000", {"name": "Luke Skywalker"}),
        ("1002", {"name": "Han Solo"}),
        ("not a valid id", None),
    ],
)
def test_generic_query_using_id_and_variable(starwars_schema, id_, expected):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query FetchSomeIDQuery($someId: String!) {
                human(id: $someId) {
                name
                }
            }
            """
        ),
        variables={"someId": id_},
    ).result()
    assert result == {"human": expected}
    assert errors == []


def test_changing_key_with_alias(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query FetchLukeAliased {
                luke: human(id: "1000") {
                name
                }
            }
            """
        ),
    ).result()
    assert result == {"luke": {"name": "Luke Skywalker"}}
    assert errors == []


def test_same_root_field_multiple_aliases(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query FetchLukeAndLeiaAliased {
                luke: human(id: "1000") {
                    name
                    homePlanet
                }
                leia: human(id: "1003") {
                    name
                    homePlanet
                }
            }
            """
        ),
    ).result()
    assert result == {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
        "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
    }
    assert errors == []


def test_use_of_fragment_to_avoid_duplicate_content(starwars_schema):
    result, errors = execute(
        starwars_schema,
        parse(
            """
            query FetchLukeAndLeiaAliased {
                luke: human(id: "1000") { ...HumanFragment }
                leia: human(id: "1003") { ...HumanFragment }
            }

            fragment HumanFragment on Human { name homePlanet }
            """
        ),
    ).result()
    assert result == {
        "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
        "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
    }
    assert errors == []


@pytest.mark.parametrize(
    "query, result",
    [
        (
            "query CheckTypeOfR2 { hero { __typename name } }",
            {"hero": {"__typename": "Droid", "name": "R2-D2"}},
        ),
        (
            "query CheckTypeOfLuke { hero(episode: EMPIRE) { __typename name } }",
            {"hero": {"__typename": "Human", "name": "Luke Skywalker"}},
        ),
    ],
)
def test_introspection(starwars_schema, query, result):
    data, errors = execute(starwars_schema, parse(query)).result()
    assert data, errors == (result, [])


def test_error_on_accessing_secret_backstory(starwars_schema):
    data, errors = execute(
        starwars_schema,
        parse(
            """
            query HeroNameQuery {
                hero {
                    name
                    secretBackstory
                }
            }
            """
        ),
    ).result()
    assert data == {"hero": {"name": "R2-D2", "secretBackstory": None}}
    assert len(errors) == 1
    err = errors[0]
    assert str(err) == "secretBackstory is secret."
    assert err.path == ["hero", "secretBackstory"]
    assert err.nodes[0].loc == (103, 118)


def test_error_on_accessing_secret_backstory_in_a_list(starwars_schema):
    data, errors = execute(
        starwars_schema,
        parse(
            """
            query HeroNameQuery {
                hero {
                    name
                    friends {
                        name
                        secretBackstory
                    }
                }
            }
            """
        ),
    ).result()
    assert data == {
        "hero": {
            "friends": [
                {"name": "Luke Skywalker", "secretBackstory": None},
                {"name": "Han Solo", "secretBackstory": None},
                {"name": "Leia Organa", "secretBackstory": None},
            ],
            "name": "R2-D2",
        }
    }
    assert len(errors) == 3
    for i, err in enumerate(errors):
        assert str(err) == "secretBackstory is secret."
        assert err.path == ["hero", "friends", i, "secretBackstory"]
        assert err.nodes[0].loc == (166, 181)


def test_error_on_accessing_secret_backstory_through_alias(starwars_schema):
    data, errors = execute(
        starwars_schema,
        parse(
            """
            query HeroNameQuery {
                mainHero: hero {
                    name
                    story: secretBackstory
                }
            }
            """
        ),
    ).result()
    assert data == {"mainHero": {"name": "R2-D2", "story": None}}
    assert len(errors) == 1
    err = errors[0]
    assert str(err) == "secretBackstory is secret."
    assert err.path == ["mainHero", "story"]
    assert err.nodes[0].loc == (113, 135)


def test_error_on_missing_argument(starwars_schema):
    data, errors = execute(
        starwars_schema,
        parse(
            """
            {
                luke: human {
                    name
                }
            }
            """
        ),
    ).result()
    assert data == {"luke": None}
    assert len(errors) == 1
    err = errors[0]
    assert (
        str(err) == 'Argument "id" of required type "String!" was not provided'
    )
    assert err.path == ["luke"]



# -*- coding: utf-8 -*-
""" Support for various runtime executors based on the :mod:`concurrent.futures`
API.

An executor is simply a subclass of :class:`concurrent.futures.Executor` which
is re-exported as :class:`~py_gql.execution.executors.Executor`.
"""
# TODO: Implement more executor options
from concurrent import futures as _f

from ._concurrency import DummyFuture

Executor = _f.Executor


class SyncExecutor(Executor):
    """ This is the default executor class which executes everything in a
    blocking manner. """

    # This tells the execution runtime which Future class to use when wrapping
    # values / consolidating multiple futures.
    # This one is much more lightweight then the standard Future and helps
    # alleviates the performance hit of using the executor / future abstractions
    # when everything is synchronous anyways.
    # REVIEW: This is awkward.
    Future = DummyFuture

    def submit(self, func, *args, **kwargs):
        return func(*args, **kwargs)

    def shutdown(self, wait=True):
        return True


DefaultExecutor = SyncExecutor
ThreadPoolExecutor = _f.ThreadPoolExecutor
